# -*- coding: utf-8 -*-
"""
    emmett.tools.auth.exposer
    -------------------------

    Provides the routes layer for the auth system.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

from typing import Any, List, Optional

from ...app import App, AppModule
from ...cache import RouteCacheRule
from ...helpers import flash, stream_dbfile
from ...http import redirect
from ...locals import request, session
from ...pipeline import Injector, Pipe, RequirePipe
from ...routing.urls import url


class AuthModule(AppModule):
    def __init__(
        self,
        app: App,
        name: str,
        import_name: str,
        template_folder: Optional[str] = None,
        template_path: Optional[str] = None,
        static_folder: Optional[str] = None,
        static_path: Optional[str] = None,
        url_prefix: Optional[str] = None,
        hostname: Optional[str] = None,
        cache: Optional[RouteCacheRule] = None,
        root_path: Optional[str] = None,
        pipeline: Optional[List[Pipe]] = None,
        injectors: Optional[List[Injector]] = None,
        **kwargs: Any
    ):
        super().__init__(
            app=app,
            name=name,
            import_name=import_name,
            template_folder=template_folder,
            template_path=template_path,
            static_folder=static_folder,
            static_path=static_path,
            url_prefix=url_prefix,
            hostname=hostname,
            cache=cache,
            root_path=root_path,
            pipeline=pipeline,
            injectors=injectors,
            **kwargs
        )
        self.init()

    def init(self):
        self.ext = self.app.ext.AuthExtension
        self.auth = self.ext.auth
        self.config = self.ext.config
        self._callbacks = {
            'after_login': self._after_login,
            'after_logout': self._after_logout,
            'after_registration': self._after_registration,
            'after_profile': self._after_profile,
            'after_email_verification': self._after_email_verification,
            'after_password_retrieval': self._after_password_retrieval,
            'after_password_reset': self._after_password_reset,
            'after_password_change': self._after_password_change
        }
        auth_pipe = [] if not self.config.inject_pipe else [self.auth.pipe]
        requires_login = [
            RequirePipe(
                lambda: self.auth.is_logged(),
                lambda: redirect(self.url('login')))]
        self._methods_pipelines = {
            'login': [],
            'logout': auth_pipe + requires_login,
            'registration': [],
            'profile': auth_pipe + requires_login,
            'email_verification': [],
            'password_retrieval': [],
            'password_reset': [],
            'password_change': auth_pipe + requires_login,
            'download': []
            # 'not_authorized': []
        }
        self.enabled_routes = list(self.config.enabled_routes)
        for method_name in self.config.disabled_routes:
            self.enabled_routes.remove(method_name)
        for key in self.enabled_routes:
            wrapper = getattr(self, key)
            f = getattr(self, "_" + key)
            wrapper()(f)
        #: configure message flashing
        if self.config.flash_messages:
            self.flash = self._flash
        else:
            self.flash = lambda *args, **kwargs: None
        #: register exposer to extensions
        self.ext.bind_exposer(self)

    def _template_for(self, key):
        tname = 'auth' if self.config.single_template else key
        return '{}{}'.format(tname, self.app.template_default_extension)

    def url(self, path, *args, **kwargs):
        path = "{}.{}".format(self.name, path)
        return url(path, *args, **kwargs)

    def _flash(self, message):
        return flash(message, 'auth')

    #: routes
    async def _login(self):
        def _validate_form(form):
            row = self.config.models['user'].get(email=form.params.email)
            if row:
                #: verify password
                if form.params.password == row.password:
                    res['user'] = row
                    return
            form.errors.email = self.config.messages['login_invalid']

        rv = {'message': None}
        res = {}
        rv['form'] = await self.ext.forms.login(onvalidation=_validate_form)
        if rv['form'].accepted:
            messages = self.config.messages
            if res['user'].registration_key == 'pending':
                rv['message'] = messages['approval_pending']
            elif res['user'].registration_key in ('disabled', 'blocked'):
                rv['message'] = messages['login_disabled']
            elif (
                res['user'].registration_key is not None and
                res['user'].registration_key.strip()
            ):
                rv['message'] = messages['verification_pending']
            if rv['message']:
                self.flash(rv['message'])
            else:
                self.ext.login_user(
                    res['user'], rv['form'].params.get('remember', False))
                self.ext.log_event(
                    self.config.messages['login_log'], {'id': res['user'].id})
                redirect_after = (await request.body_params)._after
                if redirect_after:
                    redirect(redirect_after)
                self._callbacks['after_login'](rv['form'])
        return rv

    async def _logout(self):
        self.ext.log_event(
            self.config.messages['logout_log'], {'id': self.auth.user.id})
        session.auth = None
        self.flash(self.config.messages['logged_out'])
        redirect_after = request.query_params._after
        if redirect_after:
            redirect(redirect_after)
        self._callbacks['after_logout']()

    async def _registration(self):
        def _validate_form(form):
            if form.params.password.password != form.params.password2.password:
                form.errors.password = "password mismatch"
                form.errors.password2 = "password mismatch"
                return
            del form.params.password2
            res['id'] = self.config.models['user'].table.insert(
                **form.params)

        rv = {'message': None}
        res = {}
        rv['form'] = await self.ext.forms.registration(
            onvalidation=_validate_form)
        if rv['form'].accepted:
            logged_in = False
            row = self.config.models['user'].get(res['id'])
            if self.config.registration_verification:
                email_data = {
                    'link': self.url(
                        'email_verification', row.registration_key,
                        scheme=True)}
                if not self.ext.mails['registration'](row, email_data):
                    rv['message'] = self.config.messages['mail_failure']
                    self.ext.db.rollback()
                    self.flash(rv['message'])
                    return rv
                rv['message'] = self.config.messages['mail_success']
                self.flash(rv['message'])
            elif self.config.registration_approval:
                rv['message'] = self.config.messages['approval_pending']
                self.flash(rv['message'])
            else:
                rv['message'] = self.config.messages['registration_success']
                self.flash(rv['message'])
                self.ext.login_user(row)
                logged_in = True
                self.ext.log_event(
                    self.config.messages['registration_log'],
                    {'id': res['id']})
            redirect_after = (await request.body_params)._after
            if redirect_after:
                redirect(redirect_after)
            self._callbacks['after_registration'](rv['form'], row, logged_in)
        return rv

    async def _profile(self):
        rv = {'message': None, 'form': await self.ext.forms.profile()}
        if rv['form'].accepted:
            self.auth.user.update(
                self.config.models['user'].table._filter_fields(
                    rv['form'].params))
            rv['message'] = self.config.messages['profile_updated']
            self.flash(rv['message'])
            self.ext.log_event(
                self.config.messages['profile_log'], {'id': self.auth.user.id})
            redirect_after = (await request.body_params)._after
            if redirect_after:
                redirect(redirect_after)
            self._callbacks['after_profile'](rv['form'])
        return rv

    async def _email_verification(self, key):
        rv = {'message': None}
        user = self.config.models['user'].get(registration_key=key)
        if not user:
            redirect(self.url('login'))
        if self.config.registration_approval:
            user.update_record(registration_key='pending')
            rv['message'] = self.config.messages['approval_pending']
            self.flash(rv['message'])
        else:
            user.update_record(registration_key='')
            rv['message'] = self.config.messages['verification_success']
            self.flash(rv['message'])
        #: make sure session has same user.registration_key as db record
        if self.auth.user:
            self.auth.user.registration_key = user.registration_key
        self.ext.log_event(
            self.config.messages['email_verification_log'], {'id': user.id})
        redirect_after = request.query_params._after
        if redirect_after:
            redirect(redirect_after)
        self._callbacks['after_email_verification'](user)
        return rv

    async def _password_retrieval(self):
        def _validate_form(form):
            messages = self.config.messages
            row = self.config.models['user'].get(email=form.params.email)
            if not row:
                form.errors.email = "invalid email"
                return
            if row.registration_key == 'pending':
                form.errors.email = messages['approval_pending']
                return
            if row.registration_key == 'blocked':
                form.errors.email = messages['login_disabled']
                return
            res['user'] = row

        rv = {'message': None}
        res = {}
        rv['form'] = await self.ext.forms.password_retrieval(
            onvalidation=_validate_form)
        if rv['form'].accepted:
            user = res['user']
            reset_key = self.ext.generate_reset_key(user)
            email_data = {
                'link': self.url(
                    'password_reset', reset_key, scheme=True)}
            if not self.ext.mails['reset_password'](user, email_data):
                rv['message'] = self.config.messages['mail_failure']
            rv['message'] = self.config.messages['mail_success']
            self.flash(rv['message'])
            self.ext.log_event(
                self.config.messages['password_retrieval_log'],
                {'id': user.id},
                user=user)
            redirect_after = (await request.body_params)._after
            if redirect_after:
                redirect(redirect_after)
            self._callbacks['after_password_retrieval'](user)
        return rv

    async def _password_reset(self, key):
        def _validate_form(form):
            if form.params.password.password != form.params.password2.password:
                form.errors.password = "password mismatch"
                form.errors.password2 = "password mismatch"

        rv = {'message': None}
        redirect_after = request.query_params._after
        user = self.ext.get_user_by_reset_key(key)
        if not user:
            rv['message'] = self.config.messages['reset_key_invalid']
            self.flash(rv['message'])
            if redirect_after:
                redirect(redirect_after)
            self._callbacks['after_password_reset'](user)
            return rv
        rv['form'] = await self.ext.forms.password_reset(
            onvalidation=_validate_form)
        if rv['form'].accepted:
            user.update_record(
                password=str(rv['form'].params.password),
                registration_key='',
                reset_password_key=''
            )
            rv['message'] = self.config.messages['password_changed']
            self.flash(rv['message'])
            self.ext.log_event(
                self.config.messages['password_reset_log'],
                {'id': user.id},
                user=user)
            if redirect_after:
                redirect(redirect_after)
            self._callbacks['after_password_reset'](user)
        return rv

    async def _password_change(self):
        def _validate_form(form):
            messages = self.config.messages
            if form.params.old_password != row.password:
                form.errors.old_password = messages['invalid_password']
                return
            if (
                form.params.new_password.password !=
                form.params.new_password2.password
            ):
                form.errors.new_password = "password mismatch"
                form.errors.new_password2 = "password mismatch"

        rv = {'message': None}
        row = self.config.models['user'].get(self.auth.user.id)
        rv['form'] = await self.ext.forms.password_change(
            onvalidation=_validate_form)
        if rv['form'].accepted:
            row.update_record(password=str(rv['form'].params.new_password))
            rv['message'] = self.config.messages['password_changed']
            self.flash(rv['message'])
            self.ext.log_event(
                self.config.messages['password_change_log'],
                {'id': row.id})
            redirect_after = request.query_params._after
            if redirect_after:
                redirect(redirect_after)
            self._callbacks['after_password_change']()
        return rv

    def _download(self, file_name):
        stream_dbfile(self.ext.db, file_name)

    #: routes decorators
    def login(self, template=None, pipeline=None, injectors=None):
        pipeline = self._methods_pipelines['login'] + (pipeline or [])
        return self.route(
            self.config['routes_paths']['login'], name='login',
            template=template or self._template_for('login'),
            pipeline=pipeline, injectors=injectors or [])

    def logout(self, template=None, pipeline=None, injectors=None):
        pipeline = self._methods_pipelines['logout'] + (pipeline or [])
        return self.route(
            self.config['routes_paths']['logout'], name='logout',
            template=template or self._template_for('logout'),
            pipeline=pipeline, injectors=injectors or [], methods='get')

    def registration(self, template=None, pipeline=None, injectors=None):
        pipeline = self._methods_pipelines['registration'] + (pipeline or [])
        return self.route(
            self.config['routes_paths']['registration'], name='registration',
            template=template or self._template_for('registration'),
            pipeline=pipeline, injectors=injectors or [])

    def profile(self, template=None, pipeline=None, injectors=None):
        pipeline = self._methods_pipelines['profile'] + (pipeline or [])
        return self.route(
            self.config['routes_paths']['profile'], name='profile',
            template=template or self._template_for('profile'),
            pipeline=pipeline, injectors=injectors or [])

    def email_verification(self, template=None, pipeline=None, injectors=None):
        pipeline = self._methods_pipelines['email_verification'] + (pipeline or [])
        return self.route(
            self.config['routes_paths']['email_verification'],
            name='email_verification',
            template=template or self._template_for('email_verification'),
            pipeline=pipeline, injectors=injectors or [], methods='get')

    def password_retrieval(self, template=None, pipeline=None, injectors=None):
        pipeline = self._methods_pipelines['password_retrieval'] + (pipeline or [])
        return self.route(
            self.config['routes_paths']['password_retrieval'],
            name='password_retrieval',
            template=template or self._template_for('password_retrieval'),
            pipeline=pipeline, injectors=injectors or [])

    def password_reset(self, template=None, pipeline=None, injectors=None):
        pipeline = self._methods_pipelines['password_reset'] + (pipeline or [])
        return self.route(
            self.config['routes_paths']['password_reset'],
            name='password_reset',
            template=template or self._template_for('password_reset'),
            pipeline=pipeline, injectors=injectors or [])

    def password_change(self, template=None, pipeline=None, injectors=None):
        pipeline = self._methods_pipelines['password_change'] + (pipeline or [])
        return self.route(
            self.config['routes_paths']['password_change'],
            name='password_change',
            template=template or self._template_for('password_change'),
            pipeline=pipeline, injectors=injectors or [])

    def download(self, template=None, pipeline=None, injectors=None):
        pipeline = self._methods_pipelines['download'] + (pipeline or [])
        return self.route(
            self.config['routes_paths']['download'], name='download',
            pipeline=pipeline, injectors=injectors or [], methods='get')

    #: callbacks
    def _after_login(self, form):
        redirect(self.url("profile"))

    def _after_logout(self):
        redirect(self.url("login"))

    def _after_registration(self, form, user, logged_in):
        if logged_in:
            redirect(self.url("profile"))
        redirect(self.url("login"))

    def _after_profile(self, form):
        redirect(self.url("profile"))

    def _after_email_verification(self, user):
        redirect(self.url("login"))

    def _after_password_retrieval(self, user):
        redirect(self.url("password_retrieval"))

    def _after_password_reset(self, user):
        redirect(self.url("login"))

    def _after_password_change(self):
        redirect(self.url("profile"))

    #: callbacks decorators
    def after_login(self, f):
        self._callbacks['after_login'] = f
        return f

    def after_logout(self, f):
        self._callbacks['after_logout'] = f
        return f

    def after_registration(self, f):
        self._callbacks['after_registration'] = f
        return f

    def after_profile(self, f):
        self._callbacks['after_profile'] = f
        return f

    def after_email_verification(self, f):
        self._callbacks['after_email_verification'] = f
        return f

    def after_password_retrieval(self, f):
        self._callbacks['after_password_retrieval'] = f
        return f

    def after_password_reset(self, f):
        self._callbacks['after_password_reset'] = f
        return f

    def after_password_change(self, f):
        self._callbacks['after_password_change'] = f
        return f



# -*- coding: utf-8 -*-
"""
    emmett.tools.auth.forms
    -----------------------

    Provides the forms for the authorization system.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from ...forms import Form, ModelForm
from ...orm import Field


class AuthForms(object):
    _registry_ = {}
    _fields_registry_ = {}

    @classmethod
    def register_for(cls, target, fields=lambda: None):
        def wrap(f):
            cls._registry_[target] = f
            cls._fields_registry_[target] = fields
            return f
        return wrap

    @classmethod
    def get_for(cls, target):
        return cls._registry_[target], cls._fields_registry_[target]

    @classmethod
    def map(cls):
        rv = {}
        for key in cls._registry_.keys():
            rv[key] = cls.get_for(key)
        return rv


def login_fields(auth):
    model = auth.models['user']
    rv = {
        'email': Field(
            validation={'is': 'email', 'presence': True},
            label=model.email.label),
        'password': Field(
            'password', validation=model.password._requires,
            label=model.password.label)
    }
    if auth.ext.config.remember_option:
        rv['remember'] = Field(
            'bool', default=True,
            label=auth.ext.config.messages['remember_button'])
    return rv


def registration_fields(auth):
    rw_data = auth.models['user']._instance_()._merged_form_rw_
    user_table = auth.models['user'].table
    all_fields = [
        (field_name, user_table[field_name].clone())
        for field_name in rw_data['registration']['writable']
    ]
    for i, (field_name, field) in enumerate(all_fields):
        if field_name == 'password':
            all_fields.insert(
                i + 1, (
                    'password2',
                    Field(
                        'password',
                        label=auth.ext.config.messages['verify_password'])))
            break
    rv = {}
    for i, (field_name, field) in enumerate(all_fields):
        field.writable = True
        rv[field_name] = field
        rv[field_name]._inst_count_ = i
    return rv


def profile_fields(auth):
    rw_data = auth.models['user']._instance_()._merged_form_rw_
    return rw_data['profile']


def password_retrieval_fields(auth):
    rv = {
        'email': Field(
            validation={'is': 'email', 'presence': True, 'lower': True}),
    }
    return rv


def password_reset_fields(auth):
    password_field = auth.ext.config.models['user'].password
    rv = {
        'password': Field(
            'password', validation=password_field._requires,
            label=auth.ext.config.messages['new_password']),
        'password2': Field(
            'password', label=auth.ext.config.messages['verify_password'])
    }
    return rv


def password_change_fields(auth):
    password_validation = auth.ext.config.models['user'].password._requires
    rv = {
        'old_password': Field(
            'password', validation=password_validation,
            label=auth.ext.config.messages['old_password']),
        'new_password': Field(
            'password', validation=password_validation,
            label=auth.ext.config.messages['new_password']),
        'new_password2': Field(
            'password', label=auth.ext.config.messages['verify_password'])
    }
    return rv


@AuthForms.register_for('login', fields=login_fields)
def login_form(auth, fields, **kwargs):
    opts = {
        'submit': auth.ext.config.messages['login_button'], 'keepvalues': True}
    opts.update(**kwargs)
    return Form(
        fields,
        **opts
    )


@AuthForms.register_for('registration', fields=registration_fields)
def registration_form(auth, fields, **kwargs):
    opts = {
        'submit': auth.ext.config.messages['registration_button'],
        'keepvalues': True}
    opts.update(**kwargs)
    return Form(
        fields,
        **opts
    )


@AuthForms.register_for('profile', fields=profile_fields)
def profile_form(auth, fields, **kwargs):
    opts = {
        'submit': auth.ext.config.messages['profile_button'],
        'keepvalues': True}
    opts.update(**kwargs)
    return ModelForm(
        auth.models['user'].table,
        record=auth.user,
        fields=fields,
        upload=auth.ext.exposer.url('download'),
        **opts
    )


@AuthForms.register_for('password_retrieval', fields=password_retrieval_fields)
def password_retrieval_form(auth, fields, **kwargs):
    opts = {'submit': auth.ext.config.messages['password_retrieval_button']}
    opts.update(**kwargs)
    return Form(
        fields,
        **opts
    )


@AuthForms.register_for('password_reset', fields=password_reset_fields)
def password_reset_form(auth, fields, **kwargs):
    opts = {
        'submit': auth.ext.config.messages['password_reset_button'],
        'keepvalues': True}
    opts.update(**kwargs)
    return Form(
        fields,
        **opts
    )


@AuthForms.register_for('password_change', fields=password_change_fields)
def password_change_form(auth, fields, **kwargs):
    opts = {
        'submit': auth.ext.config.messages['password_change_button'],
        'keepvalues': True}
    opts.update(**kwargs)
    return Form(
        fields,
        **opts
    )



# -*- coding: utf-8 -*-
"""
    emmett.orm.connection
    ---------------------

    Provides pyDAL connection implementation for Emmett.

    :copyright: 2014 Giovanni Barillari

    Parts of this code are inspired to peewee
    :copyright: (c) 2010 by Charles Leifer

    :license: BSD-3-Clause
"""

import asyncio
import contextvars
import heapq
import threading
import time

from collections import OrderedDict
from functools import partial

from ..ctx import current
from ..utils import cachedprop
from .errors import MaxConnectionsExceeded
from .transactions import _transaction


class ConnectionStateCtxVars:
    __slots__ = ('_connection', '_transactions', '_cursors', '_closed')

    def __init__(self):
        self._connection = contextvars.ContextVar('_emt_orm_cs_connection')
        self._transactions = contextvars.ContextVar('_emt_orm_cs_transactions')
        self._cursors = contextvars.ContextVar('_emt_orm_cs_cursors')
        self._closed = contextvars.ContextVar('_emt_orm_cs_closed')
        self.reset()

    @property
    def connection(self):
        return self._connection.get()

    @property
    def transactions(self):
        return self._transactions.get()

    @property
    def cursors(self):
        return self._cursors.get()

    @property
    def closed(self):
        return self._closed.get()

    def __set(self, connection, closed):
        self._connection.set(connection)
        self._transactions.set([])
        self._cursors.set(OrderedDict())
        self._closed.set(closed)

    def set_connection(self, connection):
        self.__set(connection, False)

    def reset(self):
        self.__set(None, True)


class ConnectionState:
    __slots__ = ('_connection', '_transactions', '_cursors', '_closed')

    def __init__(self, connection=None):
        self.connection = connection
        self._transactions = []
        self._cursors = OrderedDict()

    @property
    def connection(self):
        return self._connection

    @connection.setter
    def connection(self, value):
        self._connection = value
        self._closed = not bool(value)


class ConnectionStateCtl:
    __slots__ = ['_state_obj_var', '_state_load_var']

    state_cls = ConnectionState

    def __init__(self):
        inst_id = id(self)
        self._state_obj_var = f'__emt_orm_state_{inst_id}__'
        self._state_load_var = f'__emt_orm_state_loaded_{inst_id}__'

    @property
    def _has_ctx(self):
        return getattr(current, self._state_load_var, False)

    @property
    def ctx(self):
        if not self._has_ctx:
            setattr(current, self._state_obj_var, self.__class__.state_cls())
            setattr(current, self._state_load_var, True)
        return getattr(current, self._state_obj_var)

    @property
    def connection(self):
        return self.ctx.connection

    @property
    def transactions(self):
        return self.ctx._transactions

    @property
    def cursors(self):
        return self.ctx._cursors

    @property
    def closed(self):
        return self.ctx._closed

    def set_connection(self, connection):
        self.ctx.connection = connection

    def reset(self):
        self.ctx.connection = None
        self.ctx._transactions = []
        self.ctx._cursors = OrderedDict()


class ConnectionManager:
    __slots__ = ['adapter', 'state', '__dict__']
    state_cls = ConnectionStateCtl

    def __init__(self, adapter, **kwargs):
        self.adapter = adapter
        self.state = self.__class__.state_cls()

    def configure(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    @cachedprop
    def _loop(self):
        return asyncio.get_running_loop()

    def _connector_sync(self):
        return self.adapter.connector()

    _connector_loop = _connector_sync

    def _connection_open_sync(self):
        return self._connector_sync(), True

    async def _connection_open_loop(self):
        return (
            await self._loop.run_in_executor(
                None, self._connector_loop
            ),
            True
        )

    def _connection_close_sync(self, connection, *args, **kwargs):
        try:
            connection.close()
        except Exception:
            pass

    async def _connection_close_loop(self, connection, *args, **kwargs):
        return await self._loop.run_in_executor(
            None, partial(self._connection_close_sync, connection)
        )

    connect_sync = _connection_open_sync
    connect_loop = _connection_open_loop

    disconnect_sync = _connection_close_sync
    disconnect_loop = _connection_close_loop

    def __del__(self):
        if not self.state.closed:
            self.disconnect_sync(self.state.connection)


class PooledConnectionManager(ConnectionManager):
    __slots__ = [
        'max_connections', 'connect_timeout', 'stale_timeout',
        'connections_map', 'connections_sync',
        'in_use', '_lock_sync'
    ]

    def __init__(
        self,
        adapter,
        max_connections=5,
        connect_timeout=0,
        stale_timeout=0
    ):
        super().__init__(adapter)
        self.max_connections = max(max_connections, 1)
        self.connect_timeout = connect_timeout
        self.stale_timeout = stale_timeout
        self.connections_map = {}
        self.connections_sync = []
        self.in_use = {}
        self._lock_sync = threading.RLock()

    @cachedprop
    def _lock_loop(self):
        return asyncio.Lock()

    @cachedprop
    def connections_loop(self):
        return asyncio.LifoQueue()

    def is_stale(self, timestamp):
        return (time.time() - timestamp) > self.stale_timeout

    def connect_sync(self):
        if not self.connect_timeout:
            return self._acquire_sync()
        expires = time.time() + self.connect_timeout
        while time.time() < expires:
            try:
                rv = self._acquire_sync()
            except MaxConnectionsExceeded:
                time.sleep(0.1)
            else:
                return rv
        raise MaxConnectionsExceeded()

    async def connect_loop(self):
        return await asyncio.wait_for(
            self._acquire_loop(), self.connect_timeout or None
        )

    def _acquire_sync(self):
        _opened = False
        while True:
            try:
                with self._lock_sync:
                    ts, key = heapq.heappop(self.connections_sync)
            except IndexError:
                ts = key = conn = None
                break
            else:
                if self.stale_timeout and self.is_stale(ts):
                    self._connection_close_sync(self.connections_map.pop(key))
                else:
                    conn = self.connections_map[key]
                    break
        if conn is None:
            if len(self.connections_map) >= self.max_connections:
                raise MaxConnectionsExceeded()
            conn, _opened = self._connection_open_sync()
            ts, key = time.time(), id(conn)
            self.connections_map[key] = conn
        self.in_use[key] = ts
        return conn, _opened

    async def _acquire_loop(self):
        _opened = False
        while True:
            if len(self.connections_map) < self.max_connections:
                async with self._lock_loop:
                    if len(self.connections_map) < self.max_connections:
                        conn, _opened = await self._connection_open_loop()
                        ts, key = time.time(), id(conn)
                        self.connections_map[key] = conn
                        break
            ts, key = await self.connections_loop.get()
            if self.stale_timeout and self.is_stale(ts):
                await self._connection_close_loop(
                    self.connections_map.pop(key)
                )
            else:
                conn = self.connections_map[key]
                break
        self.in_use[key] = ts
        return conn, _opened

    def disconnect_sync(self, connection, close_connection=False):
        key = id(connection)
        ts = self.in_use.pop(key)
        if close_connection:
            self.connections_map.pop(key)
            self._connection_close_sync(connection)
        else:
            if self.stale_timeout and self.is_stale(ts):
                self.connections_map.pop(key)
                self._connection_close_sync(connection)
            else:
                with self._lock_sync:
                    heapq.heappush(self.connections_sync, (ts, key))

    async def disconnect_loop(self, connection, close_connection=False):
        key = id(connection)
        ts = self.in_use.pop(key)
        if close_connection:
            self.connections_map.pop(key)
            await self._connection_close_loop(connection)
        else:
            if self.stale_timeout and self.is_stale(ts):
                self.connections_map.pop(key)
                await self._connection_close_loop(connection)
            else:
                self.connections_loop.put_nowait((ts, key))

    def disconnect_all(self):
        for connection in self.connections_map.values():
            self.disconnect_sync(connection, close_connection=True)

    def __del__(self):
        self.disconnect_all()


def _connection_init(self, *args, **kwargs):
    self._connection_manager = self._connection_manager_cls(self)


def _connect_sync(self, with_transaction=True, reuse_if_open=False):
    if not self._connection_manager.state.closed:
        if reuse_if_open:
            return False
        raise RuntimeError('Connection already opened.')
    self.connection, _opened = self._connection_manager.connect_sync()
    if _opened:
        self.after_connection_hook()
    if with_transaction:
        txn = _transaction(self)
        txn.__enter__()
    return True


async def _connect_loop(self, with_transaction=True, reuse_if_open=False):
    if not self._connection_manager.state.closed:
        if reuse_if_open:
            return False
        raise RuntimeError('Connection already opened.')
    self.connection, _opened = await self._connection_manager.connect_loop()
    if _opened:
        self.after_connection_hook()
    if with_transaction:
        txn = _transaction(self)
        txn.__enter__()
    return True


def _close_sync(self, action='commit', really=True):
    is_open = not self._connection_manager.state.closed
    if not is_open:
        return is_open
    if self.transaction_depth() == 1:
        txn = self.top_transaction()
        try:
            txn.__exit__(None, None, None)
            succeeded = True
        except Exception:
            succeeded = False
        really = not succeeded
    try:
        self._connection_manager.disconnect_sync(self.connection, really)
    finally:
        self.connection = None
    return is_open


async def _close_loop(self, action='commit', really=True):
    is_open = not self._connection_manager.state.closed
    if not is_open:
        return is_open
    if self.transaction_depth() == 1:
        txn = self.top_transaction()
        try:
            txn.__exit__(None, None, None)
            succeeded = True
        except Exception:
            succeeded = False
        really = not succeeded
    try:
        await self._connection_manager.disconnect_loop(self.connection, really)
    finally:
        self.connection = None
    return is_open


def _connection_getter(self):
    return self._connection_manager.state.connection


def _connection_setter(self, connection):
    if connection is None:
        self._connection_manager.state.reset()
        return
    self._connection_manager.state.set_connection(connection)


def _cursors_getter(self):
    return self._connection_manager.state.cursors


def _connect_and_configure(self, *args, **kwargs):
    self._connection_reconnect(*args, **kwargs)
    with self._reconnect_lock:
        if self._reconnect_mocked:
            self._configure_on_first_reconnect()
            self.reconnect = self._connection_reconnect
            self._reconnect_mocked = False



# -*- coding: utf-8 -*-
"""
    emmett.orm.transactions
    -----------------------

    Provides pyDAL advanced transactions implementation for Emmett.

    :copyright: 2014 Giovanni Barillari

    Parts of this code are inspired to peewee
    :copyright: (c) 2010 by Charles Leifer

    :license: BSD-3-Clause
"""

import uuid
from functools import wraps


class callable_context_manager(object):
    def __call__(self, fn):
        @wraps(fn)
        def inner(*args, **kwargs):
            with self:
                return fn(*args, **kwargs)
        return inner


class _atomic(callable_context_manager):
    def __init__(self, adapter):
        self.adapter = adapter

    def __enter__(self):
        if self.adapter.transaction_depth() == 0:
            self._helper = self.adapter.db.transaction()
        else:
            self._helper = self.adapter.db.savepoint()
        return self._helper.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._helper.__exit__(exc_type, exc_val, exc_tb)


class _transaction(callable_context_manager):
    def __init__(self, adapter, lock_type=None):
        self.adapter = adapter
        self._lock_type = lock_type

    def _begin(self):
        if self._lock_type:
            self.adapter.begin(self._lock_type)
        else:
            self.adapter.begin()

    def commit(self, begin=True):
        self.adapter.commit()
        if begin:
            self._begin()

    def rollback(self, begin=True):
        self.adapter.rollback()
        if begin:
            self._begin()

    def __enter__(self):
        if self.adapter.transaction_depth() == 0:
            self._begin()
        self.adapter.push_transaction(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type:
                self.rollback(False)
            elif self.adapter.transaction_depth() == 1:
                try:
                    self.commit(False)
                except Exception:
                    self.rollback(False)
                    raise
        finally:
            self.adapter.pop_transaction()


class _savepoint(callable_context_manager):
    def __init__(self, adapter, sid=None):
        self.adapter = adapter
        self.sid = sid or 's' + uuid.uuid4().hex
        self.quoted_sid = self.adapter.dialect.quote(self.sid)

    def _begin(self):
        self.adapter.execute('SAVEPOINT %s;' % self.quoted_sid)

    def commit(self, begin=True):
        self.adapter.execute('RELEASE SAVEPOINT %s;' % self.quoted_sid)
        if begin:
            self._begin()

    def rollback(self):
        self.adapter.execute('ROLLBACK TO SAVEPOINT %s;' % self.quoted_sid)

    def __enter__(self):
        self._begin()
        self.adapter.push_transaction(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type:
                self.rollback()
            else:
                try:
                    self.commit(begin=False)
                except Exception:
                    self.rollback()
                    raise
        finally:
            self.adapter.pop_transaction()



# -*- coding: utf-8 -*-
"""
    emmett.testing.urls
    -------------------

    Provides url helpers for testing suite.

    :copyright: 2014 Giovanni Barillari

    Several parts of this code comes from Werkzeug.
    :copyright: (c) 2015 by Armin Ronacher.

    :license: BSD-3-Clause
"""

import os
import re

from collections import namedtuple

# TODO: check conversions
from .._shortcuts import to_unicode
from ..datastructures import sdict


_always_safe = (
    b'abcdefghijklmnopqrstuvwxyz'
    b'ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_.-+')
_hexdigits = '0123456789ABCDEFabcdef'
_hextobyte = dict(
    ((a + b).encode(), int(a + b, 16))
    for a in _hexdigits for b in _hexdigits
)

_scheme_re = re.compile(r'^[a-zA-Z0-9+-.]+$')

_URLTuple = namedtuple(
    '_URLTuple',
    ['scheme', 'netloc', 'path', 'query', 'fragment'])


class BaseURL(_URLTuple):
    __slots__ = ()

    def replace(self, **kwargs):
        return self._replace(**kwargs)

    @property
    def host(self):
        return self._split_host()[0]

    @property
    def ascii_host(self):
        rv = self.host
        if rv is not None and isinstance(rv, str):
            try:
                rv = _encode_idna(rv)
            except UnicodeError:
                rv = rv.encode('ascii', 'ignore')
        return to_unicode(rv, 'ascii', 'ignore')

    @property
    def port(self):
        try:
            rv = int(to_unicode(self._split_host()[1]))
            if 0 <= rv <= 65535:
                return rv
        except (ValueError, TypeError):
            pass

    @property
    def auth(self):
        return self._split_netloc()[0]

    @property
    def username(self):
        rv = self._split_auth()[0]
        if rv is not None:
            return url_unquote(rv)

    @property
    def raw_username(self):
        return self._split_auth()[0]

    @property
    def password(self):
        rv = self._split_auth()[1]
        if rv is not None:
            return url_unquote(rv)

    @property
    def raw_password(self):
        return self._split_auth()[1]

    def to_url(self):
        return url_unparse(self)

    def decode_netloc(self):
        rv = _decode_idna(self.host or '')

        if ':' in rv:
            rv = '[%s]' % rv
        port = self.port
        if port is not None:
            rv = '%s:%d' % (rv, port)
        auth = ':'.join(filter(None, [
            url_unquote(
                self.raw_username or '', errors='strict', unsafe='/:%@'),
            url_unquote(
                self.raw_password or '', errors='strict', unsafe='/:%@'),
        ]))
        if auth:
            rv = '%s@%s' % (auth, rv)
        return rv

    def to_uri_tuple(self):
        return url_parse(iri_to_uri(self).encode('ascii'))

    def to_iri_tuple(self):
        return url_parse(uri_to_iri(self))

    def get_file_location(self, pathformat=None):
        if self.scheme != 'file':
            return None, None

        path = url_unquote(self.path)
        host = self.netloc or None

        if pathformat is None:
            if os.name == 'nt':
                pathformat = 'windows'
            else:
                pathformat = 'posix'

        if pathformat == 'windows':
            if path[:1] == '/' and path[1:2].isalpha() and path[2:3] in '|:':
                path = path[1:2] + ':' + path[3:]
            windows_share = path[:3] in ('\\' * 3, '/' * 3)
            import ntpath
            path = ntpath.normpath(path)
            # Windows shared drives are represented as ``\\host\\directory``.
            # That results in a URL like ``file://///host/directory``, and a
            # path like ``///host/directory``. We need to special-case this
            # because the path contains the hostname.
            if windows_share and host is None:
                parts = path.lstrip('\\').split('\\', 1)
                if len(parts) == 2:
                    host, path = parts
                else:
                    host = parts[0]
                    path = ''
        elif pathformat == 'posix':
            import posixpath
            path = posixpath.normpath(path)
        else:
            raise TypeError('Invalid path format %s' % repr(pathformat))

        if host in ('127.0.0.1', '::1', 'localhost'):
            host = None

        return host, path

    def _split_netloc(self):
        if self._at in self.netloc:
            return self.netloc.split(self._at, 1)
        return None, self.netloc

    def _split_auth(self):
        auth = self._split_netloc()[0]
        if not auth:
            return None, None
        if self._colon not in auth:
            return auth, None
        return auth.split(self._colon, 1)

    def _split_host(self):
        rv = self._split_netloc()[1]
        if not rv:
            return None, None

        if not rv.startswith(self._lbracket):
            if self._colon in rv:
                return rv.split(self._colon, 1)
            return rv, None

        idx = rv.find(self._rbracket)
        if idx < 0:
            return rv, None

        host = rv[1:idx]
        rest = rv[idx + 1:]
        if rest.startswith(self._colon):
            return host, rest[1:]
        return host, None


class URL(BaseURL):
    __slots__ = ()
    _at = '@'
    _colon = ':'
    _lbracket = '['
    _rbracket = ']'

    def __str__(self):
        return self.to_url()

    def encode_netloc(self):
        rv = self.ascii_host or ''
        if ':' in rv:
            rv = '[%s]' % rv
        port = self.port
        if port is not None:
            rv = '%s:%d' % (rv, port)
        auth = ':'.join(filter(None, [
            url_quote(self.raw_username or '', 'utf-8', 'strict', '/:%'),
            url_quote(self.raw_password or '', 'utf-8', 'strict', '/:%'),
        ]))
        if auth:
            rv = '%s@%s' % (auth, rv)
        return to_unicode(rv)

    def encode(self, charset='utf-8', errors='replace'):
        return BytesURL(
            self.scheme.encode('ascii'),
            self.encode_netloc(),
            self.path.encode(charset, errors),
            self.query.encode(charset, errors),
            self.fragment.encode(charset, errors)
        )


class BytesURL(BaseURL):
    __slots__ = ()
    _at = b'@'
    _colon = b':'
    _lbracket = b'['
    _rbracket = b']'

    def __str__(self):
        return self.to_url().decode('utf-8', 'replace')

    def encode_netloc(self):
        return self.netloc

    def decode(self, charset='utf-8', errors='replace'):
        return URL(
            self.scheme.decode('ascii'),
            self.decode_netloc(),
            self.path.decode(charset, errors),
            self.query.decode(charset, errors),
            self.fragment.decode(charset, errors)
        )


def url_quote(string, charset='utf-8', errors='strict', safe='/:', unsafe=''):
    if not isinstance(string, (str, bytes, bytearray)):
        string = str(string)
    if isinstance(string, str):
        string = string.encode(charset, errors)
    if isinstance(safe, str):
        safe = safe.encode(charset, errors)
    if isinstance(unsafe, str):
        unsafe = unsafe.encode(charset, errors)
    safe = frozenset(bytearray(safe) + _always_safe) - \
        frozenset(bytearray(unsafe))
    rv = bytearray()
    for char in bytearray(string):
        if char in safe:
            rv.append(char)
        else:
            rv.extend(('%%%02X' % char).encode('ascii'))
    return to_unicode(bytes(rv))


def _unquote_to_bytes(string, unsafe=''):
    if isinstance(string, str):
        string = string.encode('utf-8')
    if isinstance(unsafe, str):
        unsafe = unsafe.encode('utf-8')
    unsafe = frozenset(bytearray(unsafe))
    bits = iter(string.split(b'%'))
    result = bytearray(next(bits, b''))
    for item in bits:
        try:
            char = _hextobyte[item[:2]]
            if char in unsafe:
                raise KeyError()
            result.append(char)
            result.extend(item[2:])
        except KeyError:
            result.extend(b'%')
            result.extend(item)
    return bytes(result)


def url_unquote(string, charset='utf-8', errors='replace', unsafe=''):
    rv = _unquote_to_bytes(string, unsafe)
    if charset is not None:
        rv = rv.decode(charset, errors)
    return rv


def url_quote_plus(string, charset='utf-8', errors='strict', safe=''):
    return url_quote(
        string, charset, errors, safe + ' ', '+').replace(' ', '+')


def url_unquote_plus(s, charset='utf-8', errors='replace'):
    if isinstance(s, str):
        s = s.replace(u'+', u' ')
    else:
        s = s.replace(b'+', b' ')
    return url_unquote(s, charset, errors)


def url_parse(url, scheme=None, allow_fragments=True):
    #s = make_literal_wrapper(url)
    is_text_based = isinstance(url, str)
    if scheme is None:
        scheme = ''
    netloc = query = fragment = ''
    i = url.find(':')
    if i > 0 and _scheme_re.match(to_unicode(url[:i], errors='replace')):
        # make sure "iri" is not actually a port number (in which case
        # "scheme" is really part of the path)
        rest = url[i + 1:]
        if not rest or any(c not in '0123456789' for c in rest):
            # not a port number
            scheme, url = url[:i].lower(), rest

    if url[:2] == '//':
        delim = len(url)
        for c in '/?#':
            wdelim = url.find(c, 2)
            if wdelim >= 0:
                delim = min(delim, wdelim)
        netloc, url = url[2:delim], url[delim:]
        if ('[' in netloc and ']' not in netloc) or \
           (']' in netloc and '[' not in netloc):
            raise ValueError('Invalid IPv6 URL')

    if allow_fragments and '#' in url:
        url, fragment = url.split('#', 1)
    if '?' in url:
        url, query = url.split('?', 1)

    result_type = is_text_based and URL or BytesURL
    return result_type(scheme, netloc, url, query, fragment)


def url_unparse(components):
    scheme, netloc, path, query, fragment = components
    #    normalize_string_tuple(components)
    #s = make_literal_wrapper(scheme)
    url = ''

    # We generally treat file:///x and file:/x the same which is also
    # what browsers seem to do.  This also allows us to ignore a schema
    # register for netloc utilization or having to differenciate between
    # empty and missing netloc.
    if netloc or (scheme and path.startswith('/')):
        if path and path[:1] != '/':
            path = '/' + path
        url = '//' + (netloc or '') + path
    elif path:
        url += path
    if scheme:
        url = scheme + ':' + url
    if query:
        url = url + '?' + query
    if fragment:
        url = url + '#' + fragment
    return url


def _url_encode_impl(obj, charset, encode_keys, sort, key):
    iterable = sdict()
    for key, values in obj.items():
        if not isinstance(values, list):
            values = [values]
        iterable[key] = values
    if sort:
        iterable = sorted(iterable, key=key)
    for key, values in iterable.items():
        for value in values:
            if value is None:
                continue
            if not isinstance(key, bytes):
                key = str(key).encode(charset)
            if not isinstance(value, bytes):
                value = str(value).encode(charset)
            yield url_quote_plus(key) + '=' + url_quote_plus(value)


def url_encode(obj, charset='utf-8', encode_keys=False, sort=False, key=None,
               separator=b'&'):
    separator = to_unicode(separator, 'ascii')
    return separator.join(
        _url_encode_impl(obj, charset, encode_keys, sort, key))


def uri_to_iri(uri, charset='utf-8', errors='replace'):
    if isinstance(uri, tuple):
        uri = url_unparse(uri)
    uri = url_parse(to_unicode(uri, charset))
    path = url_unquote(uri.path, charset, errors, '%/;?')
    query = url_unquote(uri.query, charset, errors, '%;/?:@&=+,$#')
    fragment = url_unquote(uri.fragment, charset, errors, '%;/?:@&=+,$#')
    return url_unparse((uri.scheme, uri.decode_netloc(),
                        path, query, fragment))


def iri_to_uri(iri, charset='utf-8', errors='strict'):
    if isinstance(iri, tuple):
        iri = url_unparse(iri)
    iri = url_parse(to_unicode(iri, charset, errors))

    netloc = iri.encode_netloc()
    path = url_quote(iri.path, charset, errors, '/:~+%')
    query = url_quote(iri.query, charset, errors, '%&[]:;$*()+,!?*/=')
    fragment = url_quote(iri.fragment, charset, errors, '=%&[]:;$()+,!?*/')

    return to_unicode(url_unparse((iri.scheme, netloc, path, query, fragment)))


def url_fix(s, charset='utf-8'):
    # First step is to switch to unicode processing and to convert
    # backslashes (which are invalid in URLs anyways) to slashes.  This is
    # consistent with what Chrome does.
    s = to_unicode(s, charset, 'replace').replace('\\', '/')

    # For the specific case that we look like a malformed windows URL
    # we want to fix this up manually:
    if (
        s.startswith('file://') and s[7:8].isalpha() and
        s[8:10] in (':/', '|/')
    ):
        s = 'file:///' + s[7:]

    url = url_parse(s)
    path = url_quote(url.path, charset, safe='/%+$!*\'(),')
    qs = url_quote_plus(url.query, charset, safe=':&%=+$!*\'(),')
    anchor = url_quote_plus(url.fragment, charset, safe=':&%=+$!*\'(),')
    return to_unicode(
        url_unparse((url.scheme, url.encode_netloc(), path, qs, anchor)))


def _encode_idna(domain):
    # If we're given bytes, make sure they fit into ASCII
    if not isinstance(domain, str):
        domain.decode('ascii')
        return domain

    # Otherwise check if it's already ascii, then return
    try:
        return domain.encode('ascii')
    except UnicodeError:
        pass

    # Otherwise encode each part separately
    parts = domain.split('.')
    for idx, part in enumerate(parts):
        parts[idx] = part.encode('idna')
    return b'.'.join(parts)


def _decode_idna(domain):
    # If the input is a string try to encode it to ascii to
    # do the idna decoding.  if that fails because of an
    # unicode error, then we already have a decoded idna domain
    if isinstance(domain, str):
        try:
            domain = domain.encode('ascii')
        except UnicodeError:
            return domain

    # Decode each part separately.  If a part fails, try to
    # decode it with ascii and silently ignore errors.  This makes
    # most sense because the idna codec does not have error handling
    parts = domain.split(b'.')
    for idx, part in enumerate(parts):
        try:
            parts[idx] = part.decode('idna')
        except UnicodeError:
            parts[idx] = part.decode('ascii', 'ignore')

    return '.'.join(parts)


def _host_is_trusted(hostname, trusted_list):
    if not hostname:
        return False

    if isinstance(trusted_list, str):
        trusted_list = [trusted_list]

    def _normalize(hostname):
        if ':' in hostname:
            hostname = hostname.rsplit(':', 1)[0]
        return _encode_idna(hostname)

    try:
        hostname = _normalize(hostname)
    except UnicodeError:
        return False
    for ref in trusted_list:
        if ref.startswith('.'):
            ref = ref[1:]
            suffix_match = True
        else:
            suffix_match = False
        try:
            ref = _normalize(ref)
        except UnicodeError:
            return False
        if ref == hostname:
            return True
        if suffix_match and hostname.endswith('.' + ref):
            return True
    return False


def get_host(scope, headers, trusted_hosts=None):
    if 'x-forwarded-host' in headers:
        rv = headers['x-forwarded-host'].split(',', 1)[0].strip()
    elif 'host' in headers:
        rv = headers['host']
    else:
        rv = scope['server'][0]
        if (
            (scope['scheme'], scope['server'][1]) not in
            (('https', '443'), ('http', '80'))
        ):
            rv += ':{}'.format(scope['server'][1])
    return rv



# -*- coding: utf-8 -*-
"""
    tests.orm
    ---------

    Test pyDAL implementation over Emmett.
"""

import pytest

from datetime import datetime, timedelta
from pydal.objects import Table
from pydal import Field as _Field
from emmett import App, sdict
from emmett.orm import (
    Database, Field, Model,
    compute,
    before_insert, after_insert,
    before_update, after_update,
    before_delete, after_delete,
    rowattr, rowmethod,
    has_one, has_many, belongs_to,
    scope
)
from emmett.validators import isntEmpty, hasLength


def _represent_f(value):
    return value


def _widget_f(field, value):
    return value


def _call_bi(fields):
    return fields[:-1]


def _call_ai(fields, id):
    return fields[:-1], id + 1


def _call_u(set, fields):
    return set, fields[:-1]


def _call_d(set):
    return set


class Stuff(Model):
    a = Field.string()
    b = Field()
    price = Field.float()
    quantity = Field.int()
    total = Field.float()
    invisible = Field()

    validation = {
        "a": {'presence': True}
    }

    fields_rw = {
        "invisible": False
    }

    form_labels = {
        "a": "A label"
    }

    form_info = {
        "a": "A comment"
    }

    update_values = {
        "a": "a_update"
    }

    repr_values = {
        "a": _represent_f
    }

    form_widgets = {
        "a": _widget_f
    }

    # def setup(self):
    #     self.table.b.requires = notInDb(self.db, self.table.b)

    @compute('total')
    def eval_total(self, row):
        return row.price * row.quantity

    @before_insert
    def bi(self, fields):
        return _call_bi(fields)

    @after_insert
    def ai(self, fields, id):
        return _call_ai(fields, id)

    @before_update
    def bu(self, set, fields):
        return _call_u(set, fields)

    @after_update
    def au(self, set, fields):
        return _call_u(set, fields)

    @before_delete
    def bd(self, set):
        return _call_d(set)

    @after_delete
    def ad(self, set):
        return _call_d(set)

    @rowattr('totalv')
    def eval_total_v(self, row):
        return row.price * row.quantity

    @rowmethod('totalm')
    def eval_total_m(self, row):
        return row.price * row.quantity

    @classmethod
    def method_test(cls, t):
        return cls.db, cls.table, t


class Person(Model):
    has_many(
        'things', {'features': {'via': 'things'}}, {'pets': 'Dog.owner'},
        'subscriptions')

    name = Field()
    age = Field.int()


class Thing(Model):
    belongs_to('person')
    has_many('features')

    name = Field()
    color = Field()


class Feature(Model):
    belongs_to('thing')
    has_one('price')

    name = Field()


class Price(Model):
    belongs_to('feature')

    value = Field.int()


class Doctor(Model):
    has_many('appointments', {'patients': {'via': 'appointments'}})
    name = Field()


class Patient(Model):
    has_many('appointments', {'doctors': {'via': 'appointments'}})
    name = Field()


class Appointment(Model):
    belongs_to('patient', 'doctor')
    date = Field.datetime()


class User(Model):
    name = Field()
    has_many(
        'memberships', {'organizations': {'via': 'memberships'}},
        {'cover_orgs': {
            'via': 'memberships.organization',
            'where': lambda m: m.is_cover == True}})


class Organization(Model):
    name = Field()
    is_cover = Field.bool(default=False)

    @has_many()
    def admin_memberships3(self):
        return Membership.admins()

    has_many(
        'memberships', {'users': {'via': 'memberships'}},
        {'admin_memberships': {'target': 'Membership', 'scope': 'admins'}},
        {'admins': {'via': 'admin_memberships.user'}},
        {'admin_memberships2': {
            'target': 'Membership', 'where': lambda m: m.role == 'admin'}},
        {'admins2': {'via': 'admin_memberships2.user'}},
        {'admins3': {'via': 'admin_memberships3.user'}})


class Membership(Model):
    belongs_to('user', 'organization')
    role = Field()

    @scope('admins')
    def filter_admins(self):
        return self.role == 'admin'


class House(Model):
    name = Field()


class Mouse(Model):
    tablename = "mice"
    has_many('elephants')
    name = Field()


class NeedSplit(Model):
    name = Field()


class Zoo(Model):
    has_many('animals', 'elephants', {'mice': {'via': 'elephants.mouse'}})
    name = Field()


class Animal(Model):
    belongs_to('zoo')
    name = Field()

    @rowattr('doublename')
    def get_double_name(self, row):
        return row.name * 2

    @rowattr('pretty')
    def get_pretty(self, row):
        return row.name

    @before_insert
    def bi(self, *args, **kwargs):
        pass

    @before_insert
    def bi2(self, *args, **kwargs):
        pass


class Elephant(Animal):
    belongs_to('mouse')
    color = Field()

    @rowattr('pretty')
    def get_pretty(self, row):
        return row.name + " " + row.color

    @before_insert
    def bi2(self, *args, **kwargs):
        pass


class Dog(Model):
    belongs_to({'owner': 'Person'})
    name = Field()


class Subscription(Model):
    belongs_to('person')

    name = Field()
    status = Field.int()
    expires_at = Field.datetime()

    STATUS = {'active': 1, 'suspended': 2, 'other': 3}

    @scope('expired')
    def get_expired(self):
        return self.expires_at < datetime.now()

    @scope('of_status')
    def filter_status(self, *statuses):
        if len(statuses) == 1:
            return self.status == self.STATUS[statuses[0]]
        return self.status.belongs(*[self.STATUS[v] for v in statuses])


@pytest.fixture(scope='module')
def db():
    app = App(__name__)
    db = Database(
        app, config=sdict(
            uri='sqlite://dal.db', auto_connect=True, auto_migrate=True))
    db.define_models([
        Stuff, Person, Thing, Feature, Price, Doctor, Patient, Appointment,
        User, Organization, Membership, House, Mouse, NeedSplit, Zoo, Animal,
        Elephant, Dog, Subscription
    ])
    return db


def test_db_instance(db):
    assert isinstance(db, Database)


def test_table_definition(db):
    assert isinstance(db.Stuff, Table)
    assert isinstance(db[Stuff.tablename], Table)


def test_fields(db):
    assert isinstance(db.Stuff.a, _Field)
    assert db.Stuff.a.type == "string"


def test_validators(db):
    assert len(db.Stuff.a.requires) == 2
    assert isinstance(db.Stuff.a.requires[0], isntEmpty)
    assert isinstance(db.Stuff.a.requires[1], hasLength)


def test_visibility(db):
    assert db.Stuff.a.readable is True
    assert db.Stuff.a.writable is True
    assert db.Stuff.invisible.readable is False
    assert db.Stuff.invisible.writable is False


def test_labels(db):
    assert db.Stuff.a.label == "A label"


def test_comments(db):
    assert db.Stuff.a.comment == "A comment"


def test_updates(db):
    assert db.Stuff.a.update == "a_update"


def test_representation(db):
    assert db.Stuff.a.represent == _represent_f


def test_widgets(db):
    assert db.Stuff.a.widget == _widget_f


# def test_set_helper(db):
#     assert isinstance(db.Stuff.b.requires, notInDb)


def test_computations(db):
    row = sdict(price=12.95, quantity=3)
    rv = db.Stuff.total.compute(row)
    assert rv == 12.95 * 3


def test_callbacks(db):
    fields = ["a", "b", "c"]
    id = 12
    rv = db.Stuff._before_insert[-1](fields)
    assert rv == fields[:-1]
    rv = db.Stuff._after_insert[-1](fields, id)
    assert rv[0] == fields[:-1] and rv[1] == id + 1
    set = {"a": "b"}
    rv = db.Stuff._before_update[-1](set, fields)
    assert rv[0] == set and rv[1] == fields[:-1]
    rv = db.Stuff._after_update[-1](set, fields)
    assert rv[0] == set and rv[1] == fields[:-1]
    rv = db.Stuff._before_delete[-1](set)
    assert rv == set
    rv = db.Stuff._after_delete[-1](set)
    assert rv == set


def test_rowattrs(db):
    db.Stuff._before_insert = []
    db.Stuff._after_insert = []
    db.Stuff.insert(a="foo", b="bar", price=12.95, quantity=3)
    db.commit()
    row = db(db.Stuff).select().first()
    assert row.totalv == 12.95 * 3


def test_rowmethods(db):
    row = db(db.Stuff).select().first()
    assert row.totalm() == 12.95 * 3


def test_modelmethods(db):
    tm = "foo"
    rv = Stuff.method_test(tm)
    assert rv[0] == db and rv[1] == db.Stuff and rv[2] == tm


def test_relations(db):
    p = db.Person.insert(name="Giovanni", age=25)
    t = db.Thing.insert(name="apple", color="red", person=p)
    f = db.Feature.insert(name="tasty", thing=t)
    db.Price.insert(value=5, feature=f)
    p = db.Person(name="Giovanni")
    #: belongs, has_one, has_many
    t = p.things()
    assert len(t) == 1
    assert t[0].name == "apple" and t[0].color == "red" and \
        t[0].person.id == p.id
    f = p.things()[0].features()
    assert len(f) == 1
    assert f[0].name == "tasty" and f[0].thing.id == t[0].id and \
        f[0].thing.person.id == p.id
    m = p.things()[0].features()[0].price()
    assert m.value == 5 and m.feature.id == f[0].id and \
        m.feature.thing.id == t[0].id and m.feature.thing.person.id == p.id
    #: has_many via as shortcut
    assert len(p.features()) == 1
    #: has_many via with join tables logic
    doctor = db.Doctor.insert(name="cox")
    patient = db.Patient.insert(name="mario")
    db.Appointment.insert(doctor=1, patient=1)
    assert len(doctor.patients()) == 1
    assert len(doctor.appointments()) == 1
    assert len(patient.doctors()) == 1
    assert len(patient.appointments()) == 1
    joe = db.User.insert(name='joe')
    jim = db.User.insert(name='jim')
    org = db.Organization.insert(name='')
    org.users.add(joe, role='admin')
    org.users.add(jim, role='manager')
    assert len(org.users()) == 2
    assert len(joe.organizations()) == 1
    assert len(jim.organizations()) == 1
    assert joe.organizations().first().id == org
    assert jim.organizations().first().id == org
    assert joe.memberships().first().role == 'admin'
    assert jim.memberships().first().role == 'manager'
    #: has_many with specified feld
    db.Dog.insert(name='pongo', owner=p)
    assert len(p.pets()) == 1 and p.pets().first().name == 'pongo'
    #: has_many via with specified field
    zoo = db.Zoo.insert(name='magic zoo')
    mouse = db.Mouse.insert(name='jerry')
    db.Elephant.insert(name='dumbo', color='pink', mouse=mouse, zoo=zoo)
    assert len(zoo.mice()) == 1


def test_tablenames(db):
    assert db.House == db.houses
    assert db.Mouse == db.mice
    assert db.NeedSplit == db.need_splits


def test_inheritance(db):
    assert 'name' in db.Animal.fields
    assert 'name' in db.Elephant.fields
    assert 'zoo' in db.Animal.fields
    assert 'zoo' in db.Elephant.fields
    assert 'color' in db.Elephant.fields
    assert 'color' not in db.Animal.fields
    assert Elephant._all_virtuals_['get_double_name'] is \
        Animal._all_virtuals_['get_double_name']
    assert Elephant._all_virtuals_['get_pretty'] is not \
        Animal._all_virtuals_['get_pretty']
    assert Elephant._all_callbacks_['bi'] is \
        Animal._all_callbacks_['bi']
    assert Elephant._all_callbacks_['bi2'] is not \
        Animal._all_callbacks_['bi2']


def test_scopes(db):
    p = db.Person.insert(name="Walter", age=50)
    s = db.Subscription.insert(
        name="a", expires_at=datetime.now() - timedelta(hours=20), person=p,
        status=1)
    s2 = db.Subscription.insert(
        name="b", expires_at=datetime.now() + timedelta(hours=20), person=p,
        status=2)
    db.Subscription.insert(
        name="c", expires_at=datetime.now() + timedelta(hours=20), person=p,
        status=3)
    rows = db(db.Subscription).expired().select()
    assert len(rows) == 1 and rows[0].id == s
    rows = p.subscriptions.expired().select()
    assert len(rows) == 1 and rows[0].id == s
    rows = Subscription.expired().select()
    assert len(rows) == 1 and rows[0].id == s
    rows = db(db.Subscription).of_status('active', 'suspended').select()
    assert len(rows) == 2 and rows[0].id == s and rows[1].id == s2
    rows = p.subscriptions.of_status('active', 'suspended').select()
    assert len(rows) == 2 and rows[0].id == s and rows[1].id == s2
    rows = Subscription.of_status('active', 'suspended').select()
    assert len(rows) == 2 and rows[0].id == s and rows[1].id == s2


def test_relations_scopes(db):
    gus = db.User.insert(name="Gus Fring")
    org = db.Organization.insert(name="Los pollos hermanos")
    org.users.add(gus, role="admin")
    frank = db.User.insert(name="Frank")
    org.users.add(frank, role='manager')
    assert org.admins.count() == 1
    assert org.admins2.count() == 1
    assert org.admins3.count() == 1
    org2 = db.Organization.insert(name="Laundry", is_cover=True)
    org2.users.add(gus, role="admin")
    assert len(gus.cover_orgs()) == 1
    assert gus.cover_orgs().first().id == org2
    org.delete_record()
    org2.delete_record()
    #: creation/addition
    org = db.Organization.insert(name="Los pollos hermanos")
    org.admins.add(gus)
    assert org.admins.count() == 1
    org.delete_record()
    org = db.Organization.insert(name="Los pollos hermanos")
    org.admins2.add(gus)
    assert org.admins2.count() == 1
    org.delete_record()
    org = db.Organization.insert(name="Los pollos hermanos")
    org.admins3.add(gus)
    assert org.admins3.count() == 1
    org.delete_record()
    gus = User.get(name="Gus Fring")
    org2 = db.Organization.insert(name="Laundry", is_cover=True)
    gus.cover_orgs.add(org2)
    assert len(gus.cover_orgs()) == 1
    assert gus.cover_orgs().first().id == org2


def test_model_where(db):
    assert Subscription.where(lambda s: s.status == 1).query == \
        db(db.Subscription.status == 1).query



# -*- coding: utf-8 -*-
"""
    emmett.tools.auth.ext
    ---------------------

    Provides the main auth layer.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import time

from functools import wraps

import click

from ...cli import pass_script_info
from ...datastructures import sdict
from ...extensions import Extension, Signals, listen_signal
from ...language.helpers import Tstr
from ...locals import T, now, session
from ...orm.helpers import decamelize
from ...security import uuid
from .forms import AuthForms
from .models import (
    AuthModel, AuthUser, AuthGroup, AuthMembership, AuthPermission, AuthEvent
)


class AuthExtension(Extension):
    namespace = 'auth'

    default_config = {
        'models': {
            'user': AuthUser,
            'group': AuthGroup,
            'membership': AuthMembership,
            'permission': AuthPermission,
            'event': AuthEvent
        },
        'hmac_key': None,
        'hmac_alg': 'pbkdf2(2000,20,sha512)',
        'inject_pipe': False,
        'log_events': True,
        'flash_messages': True,
        'csrf': True,
        'enabled_routes': [
            'login', 'logout', 'registration', 'profile', 'email_verification',
            'password_retrieval', 'password_reset', 'password_change',
            'download'],
        'disabled_routes': [],
        'routes_paths': {
            'login': '/login',
            'logout': '/logout',
            'registration': '/registration',
            'profile': '/profile',
            'email_verification': '/email_verification/<str:key>',
            'password_retrieval': '/password_retrieval',
            'password_reset': '/password_reset/<str:key>',
            'password_change': '/password_change',
            'download': '/download/<str:file_name>'
        },
        'single_template': False,
        'password_min_length': 6,
        'remember_option': True,
        'session_expiration': 3600,
        'session_long_expiration': 3600 * 24 * 30,
        'registration_verification': True,
        'registration_approval': False
    }
    default_messages = {
        'approval_pending': 'Registration is pending approval',
        'verification_pending': 'Registration needs verification',
        'login_disabled': 'The account is locked',
        'login_invalid': 'Invalid credentials',
        'logged_out': 'Logged out successfully',
        'registration_success': 'Registration completed',
        'profile_updated': 'Profile updated successfully',
        'verification_success': 'Account verification completed',
        'password_changed': 'Password changed successfully',
        'mail_failure': 'Something went wrong with the email, try again later',
        'mail_success': 'We sent you an email, check your inbox',
        'reset_key_invalid': 'The reset link was invalid or expired',
        'login_button': 'Sign in',
        'registration_button': 'Register',
        'profile_button': 'Save',
        'remember_button': 'Remember me',
        'password_retrieval_button': 'Retrieve password',
        'password_reset_button': 'Reset password',
        'password_change_button': 'Change password',
        'login_log': 'User {id} logged in',
        'logout_log': 'User {id} logged out',
        'registration_log': 'User {id} registered',
        'profile_log': 'User {id} updated profile',
        'email_verification_log': 'Verification email sent to user {id}',
        'password_retrieval_log': 'User {id} asked for password retrieval',
        'password_reset_log': 'User {id} reset the password',
        'password_change_log': 'User {id} changed the password',
        'old_password': 'Current password',
        'new_password': 'New password',
        'verify_password': 'Confirm password',
        'registration_email_subject': 'Email verification',
        'registration_email_text':
            'Hello {email}! Click on the link {link} to verify your email',
        'reset_password_email_subject': 'Password reset requested',
        'reset_password_email_text':
            'A password reset was requested for your account, '
            'click on the link {link} to proceed'
    }

    def __init__(self, app, env, config):
        super(AuthExtension, self).__init__(app, env, config)
        self.__init_messages()
        self.__init_mails()
        AuthModel._init_inheritable_dicts_()
        AuthModel.auth = self

    def __init_messages(self):
        self.config.messages = self.config.get('messages', sdict())
        for key, dval in self.default_messages.items():
            self.config.messages[key] = T(self.config.messages.get(key, dval))

    def __init_mails(self):
        self.mails = {
            'registration': self._registration_email,
            'reset_password': self._reset_password_email
        }

    def __register_commands(self):
        @self.app.cli.group('auth', short_help='Auth commands')
        def cli_group():
            pass

        @cli_group.command('generate_key', short_help='Generate an auth key')
        @pass_script_info
        def generate_key(info):
            click.echo(uuid())

    def __ensure_config(self):
        for key in (
            set(self.default_config['routes_paths'].keys()) -
            set(self.config['routes_paths'].keys())
        ):
            self.config['routes_paths'][key] = \
                self.default_config['routes_paths'][key]

    def __get_relnames(self):
        rv = {}
        def_names = {
            'user': 'user',
            'group': 'auth_group',
            'membership': 'auth_membership',
            'permission': 'auth_permission',
            'event': 'auth_event'
        }
        for m in ['user', 'group', 'membership', 'permission', 'event']:
            if self.config.models[m] == self.default_config['models'][m]:
                rv[m] = def_names[m]
            else:
                rv[m] = decamelize(self.config.models[m].__name__)
        return rv

    def on_load(self):
        self.__register_commands()
        if self.config.hmac_key is None:
            self.app.log.warn(
                "An auto-generated 'hmac_key' was added to the auth "
                "configuration.\nPlase add your own key to the configuration. "
                "You can generate a key using the auth command.\n"
                "> emmett -a {your_app_name} auth generate_key")
            self.config.hmac_key = uuid()
        self._hmac_key = self.config.hmac_alg + ':' + self.config.hmac_key
        if 'MailExtension' not in self.app.ext:
            self.app.log.warn(
                "No mailer seems to be configured. The auth features "
                "requiring mailer won't work.")
        self.__ensure_config()
        self.relation_names = self.__get_relnames()

    def bind_auth(self, auth):
        self.auth = auth

    def bind_exposer(self, exposer):
        self.exposer = exposer

    @listen_signal(Signals.before_routes)
    def _inject_pipe(self):
        if self.config.inject_pipe:
            self.app.pipeline.append(self.auth.pipe)

    def use_database(self, db, user_model=None):
        self.db = db
        if user_model:
            if not issubclass(user_model, AuthModel):
                raise RuntimeError(
                    '%s is an invalid user model' % user_model.__name__)
            self.config.models['user'] = user_model
        self.define_models()

    def __set_models_labels(self):
        for model in self.default_config['models'].values():
            for supmodel in list(reversed(model.__mro__))[1:]:
                if not supmodel.__module__.startswith(
                    'emmett.tools.auth.models'
                ):
                    continue
                if not hasattr(supmodel, 'form_labels'):
                    continue
                current_labels = {}
                for key, val in supmodel.form_labels.items():
                    current_labels[key] = T(val)
                supmodel.form_labels = current_labels

    def define_models(self):
        self.__set_models_labels()
        names = self.relation_names
        models = self.config.models
        #: AuthUser
        user_model = models['user']
        many_refs = [
            {names['membership'] + 's': models['membership'].__name__},
            {names['event'] + 's': models['event'].__name__},
            {names['group'] + 's': {'via': names['membership'] + 's'}},
            {names['permission'] + 's': {'via': names['group'] + 's'}}
        ]
        if getattr(user_model, '_auto_relations', True):
            for el in many_refs:
                key = list(el)[0]
                user_model._all_hasmany_ref_[key] = el
        if user_model.validation.get('password') is None:
            user_model.validation['password'] = {
                'len': {'gte': self.config.password_min_length},
                'crypt': {'key': self._hmac_key}
            }
        #: AuthGroup
        group_model = models['group']
        if not hasattr(group_model, 'format'):
            setattr(group_model, 'format', '%(role)s (%(id)s)')
        many_refs = [
            {names['membership'] + 's': models['membership'].__name__},
            {names['permission'] + 's': models['permission'].__name__},
            {names['user'] + 's': {'via': names['membership'] + 's'}}
        ]
        if getattr(group_model, '_auto_relations', True):
            for el in many_refs:
                key = list(el)[0]
                group_model._all_hasmany_ref_[key] = el
        #: AuthMembership
        membership_model = models['membership']
        belongs_refs = [
            {names['user']: models['user'].__name__},
            {names['group']: models['group'].__name__}
        ]
        if getattr(membership_model, '_auto_relations', True):
            for el in belongs_refs:
                key = list(el)[0]
                membership_model._all_belongs_ref_[key] = el
        #: AuthPermission
        permission_model = models['permission']
        belongs_refs = [
            {names['group']: models['group'].__name__}
        ]
        if getattr(permission_model, '_auto_relations', True):
            for el in belongs_refs:
                key = list(el)[0]
                permission_model._all_belongs_ref_[key] = el
        #: AuthEvent
        event_model = models['event']
        belongs_refs = [
            {names['user']: models['user'].__name__}
        ]
        if getattr(event_model, '_auto_relations', True):
            for el in belongs_refs:
                key = list(el)[0]
                event_model._all_belongs_ref_[key] = el
        self.db.define_models(
            models['user'], models['group'], models['membership'],
            models['permission'], models['event']
        )
        self.model_names = sdict()
        for key, value in models.items():
            self.model_names[key] = value.__name__

    def init_forms(self):
        self.forms = sdict()
        for key, (method, fields_method) in AuthForms.map().items():
            self.forms[key] = _wrap_form(
                method, fields_method(self.auth), self.auth)

    def login_user(self, user, remember=False):
        try:
            del user.password
        except Exception:
            pass
        expiration = remember and self.config.session_long_expiration or \
            self.config.session_expiration
        session.auth = sdict(
            user=user,
            last_visit=now().as_naive_datetime(),
            last_dbcheck=now().as_naive_datetime(),
            expiration=expiration,
            remember=remember
        )

    def log_event(self, description, data={}, origin='auth', user=None):
        if not self.config.log_events or not description:
            return
        try:
            user_id = user.id if user else self.auth.user.id
        except Exception:
            user_id = None
        # log messages should not be translated
        if isinstance(description, Tstr):
            description = description.text
        self.config.models['event'].table.insert(
            description=str(description % data),
            origin=origin, user=user_id)

    def generate_reset_key(self, user):
        key = str(int(time.time())) + '-' + uuid()
        user.update_record(reset_password_key=key)
        return key

    def get_user_by_reset_key(self, key):
        try:
            generated_at = int(key.split('-')[0])
            if time.time() - generated_at > 60 * 60 * 24:
                raise ValueError
            user = self.config.models['user'].get(reset_password_key=key)
        except ValueError:
            user = None
        return user

    def _registration_email(self, user, data):
        data['email'] = user.email
        return self.app.ext.MailExtension.send_mail(
            recipients=user.email,
            subject=self.config.messages['registration_email_subject'],
            body=str(self.config.messages['registration_email_text'] % data))

    def _reset_password_email(self, user, data):
        data['email'] = user.email
        return self.app.ext.MailExtension.send_mail(
            recipients=user.email,
            subject=self.config.messages['reset_password_email_subject'],
            body=str(self.config.messages['reset_password_email_text'] % data))


def _wrap_form(f, fields, auth):
    @wraps(f)
    def wrapped(*args, **kwargs):
        return f(auth, fields, *args, **kwargs)
    return wrapped



# -*- coding: utf-8 -*-
"""
    emmett.orm.engines.sqlite
    -------------------------

    Provides ORM SQLite engine specific features.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from pydal.adapters.sqlite import SQLite as _SQLite

from . import adapters


@adapters.register_for('sqlite', 'sqlite:memory')
class SQLite(_SQLite):
    def _initialize_(self, do_connect):
        super(SQLite, self)._initialize_(do_connect)
        self.driver_args['isolation_level'] = None

    def begin(self, lock_type=None):
        statement = 'BEGIN %s;' % lock_type if lock_type else 'BEGIN;'
        self.execute(statement)



# -*- coding: utf-8 -*-
"""
    emmett.wrappers.typing
    ----------------------

    Provides typing helpers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from typing import Any, Awaitable, Callable, Dict, Tuple

Scope = Dict[str, Any]
Receive = Callable[[], Awaitable[Dict[str, Any]]]
Send = Callable[[Dict[str, Any]], Awaitable[None]]
Event = Dict[str, Any]
EventHandler = Callable[[Any, Scope, Receive, Send, Event], Awaitable[Any]]
EventLooper = Callable[..., Awaitable[Tuple[EventHandler, Event]]]



# -*- coding: utf-8 -*-
"""
    emmett.language.helpers
    -----------------------

    Translation helpers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import re

from severus.datastructures import Tstr as _Tstr

from ..datastructures import Accept


class Tstr(_Tstr):
    __slots__ = []

    def __getstate__(self):
        return {
            'text': self.text,
            'lang': self.lang,
            'args': self.args,
            'kwargs': self.kwargs
        }

    def __setstate__(self, state):
        self.text = state['text']
        self.lang = state['lang']
        self.args = state['args']
        self.kwargs = state['kwargs']

    def __getattr__(self, name):
        return getattr(str(self), name)

    def __json__(self):
        return str(self)


class LanguageAccept(Accept):
    regex_locale_delim = re.compile(r'[_-]')

    def _value_matches(self, value, item):
        def _normalize(language):
            return self.regex_locale_delim.split(language.lower())[0]
        return item == '*' or _normalize(value) == _normalize(item)



# -*- coding: utf-8 -*-
"""
    tests.orm_connections
    ---------------------

    Test pyDAL connection implementation over Emmett.
"""

import pytest

from emmett import App, sdict
from emmett.orm import Database


@pytest.fixture(scope='module')
def db():
    app = App(__name__)
    db = Database(
        app,
        config=sdict(
            uri='sqlite:memory',
            auto_migrate=True,
            auto_connect=False
        )
    )
    return db


def test_connection_ctx_sync(db):
    assert not db._adapter.connection

    with db.connection():
        assert db._adapter.connection

    assert not db._adapter.connection


@pytest.mark.asyncio
async def test_connection_ctx_loop(db):
    assert not db._adapter.connection

    async with db.connection():
        assert db._adapter.connection

    assert not db._adapter.connection



# -*- coding: utf-8 -*-
"""
    emmett.orm.migrations.commands
    ------------------------------

    Provides command interfaces for migrations.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

import click

from ...datastructures import sdict
from .base import Schema, Column
from .helpers import DryRunDatabase, make_migration_id, to_tuple
from .operations import MigrationOp, UpgradeOps, DowngradeOps
from .scripts import ScriptDir


class Command(object):
    def __init__(self, app, dals):
        self.app = app
        self.envs = []
        self._load_envs(dals)

    def _load_envs(self, dals):
        for dal in dals:
            self.envs.append(
                sdict(
                    db=dal,
                    scriptdir=ScriptDir(
                        self.app, dal.config.migrations_folder
                    )
                )
            )

    def load_schema(self, ctx):
        ctx.db.define_models(Schema)
        with ctx.db.connection():
            self._ensure_schema_table_(ctx)
            self._load_current_revision_(ctx)

    def _ensure_schema_table_(self, ctx):
        # TODO -> has_table method in adapter, I don't like this "dirtness"
        try:
            ctx.db(ctx.db.Schema.id > 0).count()
        except Exception:
            ctx.db.rollback()
            from .engine import Engine
            from .operations import CreateTableOp
            op = CreateTableOp.from_table(self._build_schema_metatable_(ctx))
            op.engine = Engine(ctx.db)
            op.run()
            ctx.db.commit()

    @staticmethod
    def _build_schema_metatable_(ctx):
        from .generation import MetaTable
        columns = []
        for field in list(ctx.db.Schema):
            columns.append(Column.from_field(field))
        return MetaTable(
            ctx.db.Schema._tablename,
            columns
        )

    @staticmethod
    def _load_current_revision_(ctx):
        revisions = ctx.db(ctx.db.Schema.id > 0).select()
        if not revisions:
            ctx._current_revision_ = []
        elif len(revisions) == 1:
            ctx._current_revision_ = [revisions[0].version]
        else:
            ctx._current_revision_ = [rev.version for rev in revisions]

    @staticmethod
    def _log_store_new(revid):
        click.echo(
            " ".join([
                "> Adding revision",
                click.style(revid, fg="cyan", bold=True),
                "to schema"
            ])
        )

    @staticmethod
    def _log_store_del(revid):
        click.echo(
            " ".join([
                "> Removing revision",
                click.style(revid, fg="cyan", bold=True),
                "from schema"
            ])
        )

    @staticmethod
    def _log_store_upd(revid_src, revid_dst):
        click.echo(
            " ".join([
                "> Updating schema revision from",
                click.style(revid_src, fg="cyan", bold=True),
                "to",
                click.style(revid_dst, fg="cyan", bold=True),
            ])
        )

    @staticmethod
    def _log_dry_run(msg):
        click.secho(msg, fg='yellow')

    def _store_current_revision_(self, ctx, source, dest):
        _store_logs = {
            'new': self._log_store_new,
            'del': self._log_store_del,
            'upd': self._log_store_upd
        }
        source = to_tuple(source)
        dest = to_tuple(dest)
        if source is None:
            _store_logs['new'](dest[0])
            ctx.db.Schema.insert(version=dest[0])
            ctx.db.commit()
            ctx._current_revision_ = [dest[0]]
            return
        if dest is None:
            _store_logs['del'](source[0])
            ctx.db(ctx.db.Schema.version == source[0]).delete()
            ctx.db.commit()
            ctx._current_revision_ = []
            return
        if len(source) > 1:
            if len(source) > 2:
                ctx.db(
                    ctx.db.Schema.version.belongs(
                        source[1:])).delete()
                _store_logs['del'](source[1:])
            else:
                ctx.db(
                    ctx.db.Schema.version == source[1]).delete()
                _store_logs['del'](source[1])
            ctx.db(ctx.db.Schema.version == source[0]).update(
                version=dest[0]
            )
            _store_logs['upd'](source[0], dest[0])
            ctx._current_revision_ = [dest[0]]
        else:
            if list(source) != ctx._current_revision_:
                ctx.db.Schema.insert(version=dest[0])
                _store_logs['new'](dest[0])
                ctx._current_revision_.append(dest[0])
            else:
                ctx.db(
                    ctx.db.Schema.version == source[0]
                ).update(
                    version=dest[0]
                )
                _store_logs['upd'](source[0], dest[0])
                ctx._current_revision_ = [dest[0]]
        ctx.db.commit()

    @staticmethod
    def _generate_migration_script(ctx, migration, head):
        from .generation import Renderer
        upgrades, downgrades = Renderer.render_migration(migration)
        ctx.scriptdir.generate_revision(
            migration.rev_id, migration.message, head, upgrades=upgrades,
            downgrades=downgrades
        )

    def generate(self, message, head):
        from .generation import Generator
        for ctx in self.envs:
            upgrade_ops = Generator.generate_from(ctx.db, ctx.scriptdir, head)
            revid = make_migration_id()
            migration = MigrationOp(
                revid, upgrade_ops, upgrade_ops.reverse(), message
            )
            self._generate_migration_script(ctx, migration, head)
            click.echo(
                " ".join([
                    "> Generated migration for revision",
                    click.style(revid, fg="cyan", bold=True)
                ])
            )

    def new(self, message, head):
        for ctx in self.envs:
            source_rev = ctx.scriptdir.get_revision(head)
            revid = make_migration_id()
            migration = MigrationOp(
                revid, UpgradeOps(), DowngradeOps(), message
            )
            self._generate_migration_script(
                ctx, migration, source_rev.revision
            )
            click.echo(
                " ".join([
                    "> Created new migration for revision",
                    click.style(revid, fg="cyan", bold=True)
                ])
            )

    def history(self, base, head, verbose):
        for ctx in self.envs:
            click.echo("> Migrations history:")
            lines = []
            for sc in ctx.scriptdir.walk_revisions(
                base=base or "base",
                head=head or "heads"
            ):
                lines.append(
                    sc.cmd_format(
                        verbose=verbose, include_doc=True, include_parents=True
                    )
                )
            for line in lines:
                click.echo(line)
            if not lines:
                click.secho(
                    "No migrations for the selected application.", fg="yellow"
                )

    def status(self, verbose):
        for ctx in self.envs:
            self.load_schema(ctx)
            click.echo(
                " ".join([
                    "> Current revision(s) for",
                    click.style(ctx.db._uri, bold=True)
                ])
            )
            lines = []
            for rev in ctx.scriptdir.get_revisions(ctx._current_revision_):
                lines.append(rev.cmd_format(verbose))
            for line in lines:
                click.echo(line)
            if not lines:
                click.secho(
                    "No revision state found on the schema.", fg="yellow"
                )

    def up(self, rev_id, dry_run=False):
        log_verb = "Previewing" if dry_run else "Performing"
        for ctx in self.envs:
            self.load_schema(ctx)
            start_point = ctx._current_revision_
            revisions = ctx.scriptdir.get_upgrade_revs(
                rev_id, start_point
            )
            click.echo(
                " ".join([
                    f"> {log_verb} upgrades against",
                    click.style(ctx.db._uri, bold=True)
                ])
            )
            db = (
                DryRunDatabase(ctx.db, self._log_dry_run) if dry_run else
                ctx.db
            )
            with db.connection():
                for revision in revisions:
                    click.echo(
                        " ".join([
                            f"> {log_verb} upgrade:",
                            click.style(str(revision), fg="cyan", bold=True)
                        ])
                    )
                    migration = revision.migration_class(self.app, db)
                    try:
                        migration.up()
                        db.commit()
                        if dry_run:
                            continue
                        self._store_current_revision_(
                            ctx, migration.revises, migration.revision
                        )
                        click.echo(
                            "".join([
                                click.style(
                                    "> Succesfully upgraded to revision ",
                                    fg="green"
                                ),
                                click.style(
                                    revision.revision, fg="cyan", bold=True
                                ),
                                click.style(f": {revision.doc}", fg="green")
                            ])
                        )
                    except Exception:
                        db.rollback()
                        click.echo(
                            " ".join([
                                click.style("> Failed upgrading to", fg="red"),
                                click.style(
                                    revision.revision, fg="red", bold=True
                                ),
                            ])
                        )
                        raise

    def down(self, rev_id, dry_run=False):
        log_verb = "Previewing" if dry_run else "Performing"
        for ctx in self.envs:
            self.load_schema(ctx)
            start_point = ctx._current_revision_
            revisions = ctx.scriptdir.get_downgrade_revs(
                rev_id, start_point)
            click.echo(
                " ".join([
                    f"> {log_verb} downgrades against",
                    click.style(ctx.db._uri, bold=True)
                ])
            )
            db = (
                DryRunDatabase(ctx.db, self._log_dry_run) if dry_run else
                ctx.db
            )
            with db.connection():
                for revision in revisions:
                    click.echo(
                        " ".join([
                            f"> {log_verb} downgrade:",
                            click.style(str(revision), fg="cyan", bold=True)
                        ])
                    )
                    migration = revision.migration_class(self.app, db)
                    try:
                        migration.down()
                        db.commit()
                        if dry_run:
                            continue
                        self._store_current_revision_(
                            ctx, migration.revision, migration.revises
                        )
                        click.echo(
                            "".join([
                                click.style(
                                    "> Succesfully downgraded from revision ",
                                    fg="green"
                                ),
                                click.style(
                                    revision.revision, fg="cyan", bold=True
                                ),
                                click.style(f": {revision.doc}", fg="green")
                            ])
                        )
                    except Exception:
                        db.rollback()
                        click.echo(
                            " ".join([
                                click.style(
                                    "> Failed downgrading from", fg="red"
                                ),
                                click.style(
                                    revision.revision, fg="red", bold=True
                                ),
                            ])
                        )
                        raise


def generate(app, dals, message, head):
    Command(app, dals).generate(message, head)


def new(app, dals, message, head):
    Command(app, dals).new(message, head)


def history(app, dals, rev_range, verbose):
    if rev_range is not None:
        if ":" not in rev_range:
            raise Exception(
                "History range requires [start]:[end], "
                "[start]:, or :[end]")
        base, head = rev_range.strip().split(":")
    else:
        base = head = None
    Command(app, dals).history(base, head, verbose)


def status(app, dals, verbose):
    Command(app, dals).status(verbose)


def up(app, dals, revision, dry_run):
    Command(app, dals).up(revision, dry_run)


def down(app, dals, revision, dry_run):
    Command(app, dals).down(revision, dry_run)



# -*- coding: utf-8 -*-
"""
    tests.logger
    ------------

    Test Emmett logging module
"""

import logging

from emmett import App, logger, sdict


def _call_create_logger(app):
    return logger.create_logger(app)


def test_user_assign_valid_level():
    app = App(__name__)
    app.config.logging.pytest = sdict(
        level='info'
    )
    result = _call_create_logger(app)
    assert result.handlers[-1].level == logging.INFO


def test_user_assign_invaild_level():
    app = App(__name__)
    app.config.logging.pytest = sdict(
        level='invalid'
    )
    result = _call_create_logger(app)
    assert result.handlers[-1].level == logging.WARNING


def test_user_no_assign_level():
    app = App(__name__)
    app.config.logging.pytest = sdict()
    result = _call_create_logger(app)
    assert result.handlers[-1].level == logging.WARNING



# -*- coding: utf-8 -*-
"""
    tests.migrations
    ----------------

    Test Emmett migrations engine
"""

import pytest

from emmett import App
from emmett.orm import Database, Model, Field
from emmett.orm.migrations.engine import MetaEngine, Engine
from emmett.orm.migrations.generation import MetaData, Comparator


class FakeEngine(Engine):
    sql_history = []

    def _log_and_exec(self, sql):
        self.sql_history.append(sql)


@pytest.fixture(scope='module')
def app():
    rv = App(__name__)
    rv.config.db.uri = 'sqlite:memory'
    return rv


def _load_on_meta(ops):
    db = MetaData()
    engine = MetaEngine(db)
    for op in ops:
        op.engine = engine
        op.run()
    return db


def _make_ops(db, base=None):
    if base is not None:
        mdb = _load_on_meta(base.ops)
    else:
        mdb = MetaData()
    return Comparator.compare(db, mdb)


def _make_sql(db, op):
    engine = FakeEngine(db)
    op.engine = engine
    op.run()
    return engine.sql_history[-1] if engine.sql_history else None


class StepOneThing(Model):
    name = Field()
    value = Field.float()

_step_one_sql = """CREATE TABLE "step_one_things"(
    "id" INTEGER PRIMARY KEY AUTOINCREMENT,
    "name" CHAR(512),
    "value" DOUBLE
);"""

_step_one_sql_drop = 'DROP TABLE "step_one_things";'


def test_step_one_create_table(app):
    db = Database(app, auto_migrate=False)
    db.define_models(StepOneThing)
    ops = _make_ops(db)
    diffs = ops.as_diffs()
    assert len(diffs) == 1 and diffs[0][0] == "add_table"
    op = ops.ops[0]
    sql = _make_sql(db, op)
    assert sql == _step_one_sql


def test_step_one_no_diff_in_migration(app):
    db = Database(app, auto_migrate=False)
    db.define_models(StepOneThing)
    ops = _make_ops(db)
    ops2 = _make_ops(db, ops)
    assert len(ops2.as_diffs()) == 0


def test_step_one_drop_table(app):
    db = Database(app, auto_migrate=False)
    db.define_models(StepOneThing)
    ops = _make_ops(db)
    db2 = Database(app, auto_migrate=False)
    db2.define_models()
    ops2 = _make_ops(db2, ops)
    diffs = ops2.as_diffs()
    assert len(diffs) == 1 and diffs[0][0] == "remove_table"
    op = ops2.ops[0]
    sql = _make_sql(db2, op)
    assert sql == _step_one_sql_drop


class StepTwoThing(Model):
    name = Field(notnull=True)
    value = Field.float(default=8.8)
    available = Field.bool(default=True)

_step_two_sql = """CREATE TABLE "step_two_things"(
    "id" INTEGER PRIMARY KEY AUTOINCREMENT,
    "name" CHAR(512) NOT NULL,
    "value" DOUBLE DEFAULT '8.8',
    "available" CHAR(1) DEFAULT 'T'
);"""


def test_step_two_create_table(app):
    db = Database(app, auto_migrate=False)
    db.define_models(StepTwoThing)
    ops = _make_ops(db)
    op = ops.ops[0]
    sql = _make_sql(db, op)
    assert sql == _step_two_sql


class StepThreeThingOne(Model):
    a = Field()


class StepThreeThingTwo(Model):
    tablename = "step_three_thing_ones"
    a = Field()
    b = Field()


class StepThreeThingThree(Model):
    tablename = "step_three_thing_ones"
    b = Field()

_step_three_sql = 'ALTER TABLE "step_three_thing_ones" ADD "b" CHAR(512);'
_step_three_sql_drop = 'ALTER TABLE "step_three_thing_ones" DROP COLUMN "a";'


def test_step_three_create_column(app):
    db = Database(app, auto_migrate=False)
    db.define_models(StepThreeThingOne)
    ops = _make_ops(db)
    db2 = Database(app, auto_migrate=False)
    db2.define_models(StepThreeThingTwo)
    ops2 = _make_ops(db2, ops)
    op = ops2.ops[0]
    sql = _make_sql(db, op)
    assert sql == _step_three_sql


def test_step_three_drop_column(app):
    db = Database(app, auto_migrate=False)
    db.define_models(StepThreeThingTwo)
    ops = _make_ops(db)
    db2 = Database(app, auto_migrate=False)
    db2.define_models(StepThreeThingThree)
    ops2 = _make_ops(db2, ops)
    op = ops2.ops[0]
    sql = _make_sql(db, op)
    assert sql == _step_three_sql_drop


class StepFourThing(Model):
    name = Field(notnull=True)
    value = Field.float(default=8.8)
    available = Field.bool(default=True)
    asd = Field()


class StepFourThingEdit(Model):
    tablename = "step_four_things"
    name = Field()
    value = Field.float()
    available = Field.bool(default=True)
    asd = Field.int()

_step_four_sql = """ALTER TABLE "step_four_things" ALTER COLUMN "name" DROP NOT NULL;
ALTER TABLE "step_four_things" ALTER COLUMN "value" DROP DEFAULT;
ALTER TABLE "step_four_things" ALTER COLUMN "asd" TYPE INTEGER;"""


def test_step_four_alter_table(app):
    db = Database(app, auto_migrate=False)
    db.define_models(StepFourThing)
    ops = _make_ops(db)
    db2 = Database(app, auto_migrate=False)
    db2.define_models(StepFourThingEdit)
    ops2 = _make_ops(db2, ops)
    sql = []
    for op in ops2.ops:
        sql.append(_make_sql(db2, op))
    assert "\n".join(sql) == _step_four_sql


class StepFiveThing(Model):
    name = Field()
    value = Field.int()
    created_at = Field.datetime()

    indexes = {
        'name': True,
        ('name', 'value'): True
    }


class StepFiveThingEdit(StepFiveThing):
    tablename = "step_five_things"

    indexes = {
        'name': False,
        'name_created': {
            'fields': 'name',
            'expressions': lambda m: m.created_at.coalesce(None)}
    }


_step_five_sql_before = [
    'CREATE INDEX "step_five_things_widx__name" ON "step_five_things" ("name");',
    'CREATE INDEX "step_five_things_widx__name_value" ON "step_five_things" ("name","value");'
]

_step_five_sql_after = [
    'DROP INDEX "step_five_things_widx__name";',
    'CREATE INDEX "step_five_things_widx__name_created" ON "step_five_things" ("name",COALESCE("created_at",NULL));'
]


def test_step_five_indexes(app):
    db = Database(app, auto_migrate=False)
    db.define_models(StepFiveThing)
    ops = _make_ops(db)
    index_ops = ops.ops[1:]
    for op in index_ops:
        sql = _make_sql(db, op)
        assert sql in _step_five_sql_before
    db2 = Database(app, auto_migrate=False)
    db2.define_models(StepFiveThingEdit)
    ops2 = _make_ops(db2, ops)
    for op in ops2.ops:
        sql = _make_sql(db, op)
        assert sql in _step_five_sql_after



# -*- coding: utf-8 -*-
"""
    pbkdf2
    ~~~~~~

    This module implements pbkdf2 for Python.  It also has some basic
    tests that ensure that it works.  The implementation is straightforward
    and uses stdlib only stuff and can be easily be copy/pasted into
    your favourite application.

    Use this as replacement for bcrypt that does not need a c implementation
    of a modified blowfish crypto algo.

    Example usage:

    >>> pbkdf2_hex('what i want to hash', 'the random salt')
    'fa7cc8a2b0a932f8e6ea42f9787e9d36e592e0c222ada6a9'

    How to use this:

    1.  Use a constant time string compare function to compare the stored hash
        with the one you're generating::

            def safe_str_cmp(a, b):
                if len(a) != len(b):
                    return False
                rv = 0
                for x, y in izip(a, b):
                    rv |= ord(x) ^ ord(y)
                return rv == 0

    2.  Use `os.urandom` to generate a proper salt of at least 8 byte.
        Use a unique salt per hashed password.

    3.  Store ``algorithm$salt:costfactor$hash`` in the database so that
        you can upgrade later easily to a different algorithm if you need
        one.  For instance ``PBKDF2-256$thesalt:10000$deadbeef...``.


    :copyright: (c) Copyright 2011 by Armin Ronacher.
    :license: BSD-3-Clause
"""
import codecs
import sys
import hmac
import hashlib

from itertools import starmap
from struct import Struct
from operator import xor
from emmett._shortcuts import to_bytes

izip = zip
xrange = range
_pack_int = Struct('>I').pack


def pbkdf2_hex(data, salt, iterations=1000, keylen=24, hashfunc=None):
    """Like :func:`pbkdf2_bin` but returns a hex encoded string."""
    rv = pbkdf2_bin(data, salt, iterations, keylen, hashfunc)
    return codecs.encode(rv, 'hex_codec').decode("utf8")


def pbkdf2_bin(data, salt, iterations=1000, keylen=24, hashfunc=None):
    """Returns a binary digest for the PBKDF2 hash algorithm of `data`
    with the given `salt`.  It iterates `iterations` time and produces a
    key of `keylen` bytes.  By default SHA-1 is used as hash function,
    a different hashlib `hashfunc` can be provided.
    """
    hashfunc = hashfunc or hashlib.sha1
    mac = hmac.HMAC(data, None, hashfunc)
    if not keylen:
        keylen = mac.digest_size

    def _pseudorandom(x, mac=mac):
        h = mac.copy()
        h.update(x)
        return bytearray(h.digest())
    buf = bytearray()
    for block in xrange(1, -(-keylen // mac.digest_size) + 1):
        rv = u = _pseudorandom(salt + _pack_int(block))
        for i in xrange(iterations - 1):
            u = _pseudorandom(bytes(u))
            rv = bytearray(starmap(xor, izip(rv, u)))
        buf.extend(rv)
    return bytes(buf[:keylen])


def test():
    failed = []

    def check(data, salt, iterations, keylen, expected):
        rv = pbkdf2_hex(to_bytes(data), to_bytes(salt), iterations, keylen)
        if rv != expected:
            sys.stdout.write('Test failed:\n')
            sys.stdout.write('  Expected:   %s\n' % expected)
            sys.stdout.write('  Got:        %s\n' % rv)
            sys.stdout.write('  Parameters:\n')
            sys.stdout.write('    data=%s\n' % data)
            sys.stdout.write('    salt=%s\n' % salt)
            sys.stdout.write('    iterations=%d\n' % iterations)
            sys.stdout.write('\n')
            failed.append(1)

    # From RFC 6070
    check('password', 'salt', 1, 20,
          '0c60c80f961f0e71f3a9b524af6012062fe037a6')
    check('password', 'salt', 2, 20,
          'ea6c014dc72d6f8ccd1ed92ace1d41f0d8de8957')
    check('password', 'salt', 4096, 20,
          '4b007901b765489abead49d926f721d065a429c1')
    check('passwordPASSWORDpassword', 'saltSALTsaltSALTsaltSALTsaltSALTsalt',
          4096, 25, '3d2eec4fe41c849b80c8d83662c0e44a8b291a964cf2f07038')
    check('pass\x00word', 'sa\x00lt', 4096, 16,
          '56fa6aa75548099dcc37d7f03425e0c3')
    # This one is from the RFC but it just takes for ages
    # check('password', 'salt', 16777216, 20,
    #      'eefe3d61cd4da4e4e9945b3d6ba2158c2634e984')

    # From Crypt-PBKDF2
    check('password', 'ATHENA.MIT.EDUraeburn', 1, 16,
          'cdedb5281bb2f801565a1122b2563515')
    check('password', 'ATHENA.MIT.EDUraeburn', 1, 32,
          'cdedb5281bb2f801565a1122b25635150ad1f7a04bb9f3a333ecc0e2e1f70837')
    check('password', 'ATHENA.MIT.EDUraeburn', 2, 16,
          '01dbee7f4a9e243e988b62c73cda935d')
    check('password', 'ATHENA.MIT.EDUraeburn', 2, 32,
          '01dbee7f4a9e243e988b62c73cda935da05378b93244ec8f48a99e61ad799d86')
    check('password', 'ATHENA.MIT.EDUraeburn', 1200, 32,
          '5c08eb61fdf71e4e4ec3cf6ba1f5512ba7e52ddbc5e5142f708a31e2e62b1e13')
    check('X' * 64, 'pass phrase equals block size', 1200, 32,
          '139c30c0966bc32ba55fdbf212530ac9c5ec59f1a452f5cc9ad940fea0598ed1')
    check('X' * 65, 'pass phrase exceeds block size', 1200, 32,
          '9ccad6d468770cd51b10e6a68721be611a8b4d282601db3b36be9246915ec82a')

    raise SystemExit(bool(failed))


if __name__ == '__main__':
    test()



from . import _patches
from .adapters import adapters as adapters_registry
from .base import Database
from .objects import Field
from .models import Model
from .apis import (
    belongs_to, refers_to, has_one, has_many,
    compute, rowattr, rowmethod,
    before_insert, before_update, before_delete,
    after_insert, after_update, after_delete,
    scope
)



# -*- coding: utf-8 -*-
"""
    tests.auth
    ----------

    Test Emmett Auth module
"""

import os
import pytest
import shutil
from emmett import App
from emmett.orm import Database, Field, Model, has_many, belongs_to
from emmett.sessions import SessionManager
from emmett.tools import Auth, Mailer
from emmett.tools.auth.models import AuthUser


class User(AuthUser):
    has_many('things')
    gender = Field()


class Thing(Model):
    belongs_to('user')


@pytest.fixture(scope='module')
def app():
    rv = App(__name__)
    rv.config.mailer.sender = 'nina@massivedynamics.com'
    rv.config.auth.single_template = True
    rv.config.auth.hmac_key = "foobar"
    rv.pipeline = [SessionManager.cookies('foobar')]
    return rv


@pytest.fixture(scope='module')
def client(app):
    return app.test_client()


@pytest.fixture(scope='module')
def mailer(app):
    return Mailer(app)


@pytest.fixture(scope='module')
def _db(app):
    try:
        shutil.rmtree(os.path.join(app.root_path, 'databases'))
    except:
        pass
    db = Database(app, auto_migrate=True)
    app.pipeline.append(db.pipe)
    return db


@pytest.fixture(scope='module')
def auth(app, _db, mailer):
    auth = Auth(app, _db, user_model=User)
    app.pipeline.append(auth.pipe)
    authroutes = auth.module(__name__)
    return auth


@pytest.fixture(scope='module')
def db(_db, auth):
    with _db.connection():
        _db.define_models(Thing)
    return _db


def test_models(db):
    with db.connection():
        user = User.create(
            email="walter@massivedynamics.com",
            password="pocketuniverse",
            first_name="Walter",
            last_name="Bishop"
        )
        group = db.auth_groups.insert(
            role="admin"
        )
        group2 = db.auth_groups.insert(
            role="moderator"
        )
        db.auth_memberships.insert(
            user=user.id,
            auth_group=group
        )
        db.auth_permissions.insert(
            auth_group=group
        )
        user = db.users[1]
        assert len(user.auth_memberships()) == 1
        assert user.auth_memberships()[0].user == 1
        assert len(user.auth_groups()) == 1
        assert user.auth_groups()[0].role == "admin"
        assert len(user.auth_permissions()) == 1
        assert user.auth_permissions()[0].auth_group.role == "admin"
        assert len(group.auth_memberships()) == 1
        assert group.auth_memberships()[0].auth_group == 1
        assert len(group.users()) == 1
        assert group.users()[0].email == "walter@massivedynamics.com"
        assert len(group.auth_permissions()) == 1
        user.auth_groups.add(group2)
        assert len(user.auth_groups(db.auth_groups.id)) == 2
        assert user.auth_groups(db.auth_groups.role)[1].role == "moderator"
        assert len(user.things()) == 0


def test_registration(mailer, db, client):
    page = client.get('/auth/registration').data
    assert 'E-mail' in page
    assert 'First name' in page
    assert 'Last name' in page
    assert 'Password' in page
    assert 'Confirm password' in page
    assert 'Register' in page
    with mailer.store_mails() as mailbox:
        with client.get('/auth/registration').context as ctx:
            req = client.post('/auth/registration', data={
                'email': 'william@massivedynamics.com',
                'first_name': 'William',
                'last_name': 'Bell',
                'password': 'imtheceo',
                'password2': 'imtheceo',
                '_csrf_token': list(ctx.session._csrf)[-1]
            }, follow_redirects=True)
            assert "We sent you an email, check your inbox" in req.data
        assert len(mailbox) == 1
        mail = mailbox[0]
        assert mail.recipients == ["william@massivedynamics.com"]
        assert mail.subject == 'Email verification'
        mail_as_str = str(mail)
        assert 'Hello william@massivedynamics.com!' in mail_as_str
        assert 'verify your email' in mail_as_str
        verification_code = mail_as_str.split(
            "http://localhost/auth/email_verification/")[1].split(" ")[0]
        req = client.get(
            '/auth/email_verification/{}'.format(verification_code),
            follow_redirects=True)
        assert "Account verification completed" in req.data


def test_login(db, client):
    page = client.get('/auth/login').data
    assert 'E-mail' in page
    assert 'Password' in page
    assert 'Sign in' in page
    with client.get('/auth/login').context as ctx:
        req = client.post('/auth/login', data={
            'email': 'william@massivedynamics.com',
            'password': 'imtheceo',
            '_csrf_token': list(ctx.session._csrf)[-1]
        }, follow_redirects=True)
        assert 'William' in req.data
        assert 'Bell' in req.data
        assert 'Save' in req.data


def test_password_change(db, client):
    with client.get('/auth/login').context as ctx:
        with client.post('/auth/login', data={
            'email': 'william@massivedynamics.com',
            'password': 'imtheceo',
            '_csrf_token': list(ctx.session._csrf)[-1]
        }, follow_redirects=True):
            page = client.get('/auth/password_change').data
            assert 'Current password' in page
            assert 'New password' in page
            assert 'Confirm password' in page
            with client.get('/auth/password_change').context as ctx2:
                with client.post('/auth/password_change', data={
                    'old_password': 'imtheceo',
                    'new_password': 'imthebigceo',
                    'new_password2': 'imthebigceo',
                    '_csrf_token': list(ctx2.session._csrf)[-1]
                }, follow_redirects=True) as req:
                    assert 'Password changed successfully' in req.data
                    assert 'William' in req.data
                    assert 'Save' in req.data
    with db.connection():
        assert (
            db.users(email='william@massivedynamics.com').password ==
            db.users.password.requires[-1]('imthebigceo')[0])


def test_password_retrieval(mailer, db, client):
    page = client.get('/auth/password_retrieval').data
    assert 'Email' in page
    assert 'Retrieve password' in page
    with mailer.store_mails() as mailbox:
        with client.get('/auth/password_retrieval').context as ctx:
            with client.post('/auth/password_retrieval', data={
                'email': 'william@massivedynamics.com',
                '_csrf_token': list(ctx.session._csrf)[-1]
            }, follow_redirects=True) as req:
                assert 'We sent you an email, check your inbox' in req.data
            assert len(mailbox) == 1
        mail = mailbox[0]
        assert mail.recipients == ["william@massivedynamics.com"]
        assert mail.subject == 'Password reset requested'
        mail_as_str = str(mail)
        assert 'A password reset was requested for your account' in mail_as_str
        reset_code = mail_as_str.split(
            "http://localhost/auth/password_reset/")[1].split(" ")[0]
        with client.get(
            '/auth/password_reset/{}'.format(reset_code),
            follow_redirects=True
        ) as req:
            assert 'New password' in req.data
            assert 'Confirm password' in req.data
            assert 'Reset password' in req.data
            with client.post(
                '/auth/password_reset/{}'.format(reset_code),
                data={
                    'password': 'imtheceo',
                    'password2': 'imtheceo',
                    '_csrf_token': list(req.context.session._csrf)[-1]},
                follow_redirects=True
            ) as req2:
                assert 'Password changed successfully' in req2.data
                assert 'Sign in' in req2.data
    with db.connection():
        assert (
            db.users(email='william@massivedynamics.com').password ==
            db.users.password.requires[-1]('imtheceo')[0])



# -*- coding: utf-8 -*-
"""
    emmett.asgi.handlers
    --------------------

    Provides ASGI handlers.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from __future__ import annotations

import asyncio
import os
import re

from collections import OrderedDict
from typing import Any, Awaitable, Callable, Optional, Tuple, Union

from ..ctx import RequestContext, WSContext, current
from ..debug import smart_traceback, debug_handler
from ..http import HTTPResponse, HTTPFile, HTTP
from ..utils import cachedprop
from ..wrappers.helpers import RequestCancelled
from ..wrappers.request import Request
from ..wrappers.response import Response
from ..wrappers.websocket import Websocket
from .typing import Event, EventHandler, EventLooper, Receive, Scope, Send

REGEX_STATIC = re.compile(
    r'^/static/(?P<m>__[\w\-\.]+__/)?(?P<v>_\d+\.\d+\.\d+/)?(?P<f>.*?)$'
)
REGEX_STATIC_LANG = re.compile(
    r'^/(?P<l>\w{2}/)?static/(?P<m>__[\w\-\.]__+/)?(?P<v>_\d+\.\d+\.\d+/)?(?P<f>.*?)$'
)


class EventHandlerWrapper:
    __slots__ = ['event', 'f']

    def __init__(self, event: str, f: EventHandler):
        self.event = event
        self.f = f

    async def __call__(
        self,
        handler: Handler,
        scope: Scope,
        receive: Receive,
        send: Send,
        event: Event
    ) -> Tuple[Optional[EventHandler], None]:
        task = await self.f(handler, scope, receive, send, event)
        return task, None


class MetaHandler(type):
    def __new__(cls, name, bases, attrs):
        new_class = type.__new__(cls, name, bases, attrs)
        declared_events = OrderedDict()
        all_events = OrderedDict()
        events = []
        for key, value in list(attrs.items()):
            if isinstance(value, EventHandlerWrapper):
                events.append((key, value))
        declared_events.update(events)
        new_class._declared_events_ = declared_events
        for base in reversed(new_class.__mro__[1:]):
            if hasattr(base, '_declared_events_'):
                all_events.update(base._declared_events_)
        all_events.update(declared_events)
        new_class._all_events_ = all_events
        new_class._events_handlers_ = {
            el.event: el for el in new_class._all_events_.values()}
        return new_class


class Handler(metaclass=MetaHandler):
    __slots__ = ['app']

    def __init__(self, app):
        self.app = app

    @classmethod
    def on_event(
        cls, event: str
    ) -> Callable[[EventHandler], EventHandlerWrapper]:
        def wrap(f: EventHandler) -> EventHandlerWrapper:
            return EventHandlerWrapper(event, f)
        return wrap

    def get_event_handler(
        self, event_type: str
    ) -> Union[EventHandler, EventHandlerWrapper]:
        return self._events_handlers_.get(event_type, _event_missing)

    def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ) -> Awaitable[None]:
        return self.handle_events(scope, receive, send)

    async def handle_events(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ):
        task: Optional[EventLooper] = _event_looper
        event = None
        while task:
            task, event = await task(self, scope, receive, send, event)


class LifeSpanHandler(Handler):
    __slots__ = []

    @Handler.on_event('lifespan.startup')
    async def event_startup(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        event: Event
    ) -> EventLooper:
        await send({'type': 'lifespan.startup.complete'})
        return _event_looper

    @Handler.on_event('lifespan.shutdown')
    async def event_shutdown(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        event: Event
    ):
        await send({'type': 'lifespan.shutdown.complete'})


class RequestHandler(Handler):
    __slots__ = ['router']

    def __init__(self, app):
        super().__init__(app)
        self._bind_router()
        self._configure_methods()

    def _bind_router(self):
        raise NotImplementedError

    def _configure_methods(self):
        raise NotImplementedError


class HTTPHandler(RequestHandler):
    __slots__ = ['pre_handler', 'static_handler', 'static_matcher', '__dict__']

    def _bind_router(self):
        self.router = self.app._router_http

    def _configure_methods(self):
        self.static_matcher = (
            self._static_lang_matcher if self.app.language_force_on_url else
            self._static_nolang_matcher)
        self.static_handler = (
            self._static_handler if self.app.config.handle_static else
            self.dynamic_handler)
        self.pre_handler = (
            self._prefix_handler if self.router._prefix_main else
            self.static_handler)

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ):
        scope['emt.path'] = scope['path'] or '/'
        try:
            http = await self.pre_handler(scope, receive, send)
            await asyncio.wait_for(
                http.send(scope, send),
                self.app.config.response_timeout
            )
        except RequestCancelled:
            return
        except asyncio.TimeoutError:
            self.app.log.warn(
                f"Timeout sending response: ({scope['emt.path']})"
            )

    @cachedprop
    def error_handler(self) -> Callable[[], Awaitable[str]]:
        return (
            self._debug_handler if self.app.debug else self.exception_handler
        )

    @cachedprop
    def exception_handler(self) -> Callable[[], Awaitable[str]]:
        return self.app.error_handlers.get(500, self._exception_handler)

    @staticmethod
    async def _http_response(code: int) -> HTTPResponse:
        return HTTP(code)

    def _prefix_handler(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ) -> Awaitable[HTTPResponse]:
        path = scope['emt.path']
        if not path.startswith(self.router._prefix_main):
            return self._http_response(404)
        scope['emt.path'] = path[self.router._prefix_main_len:] or '/'
        return self.static_handler(scope, receive, send)

    def _static_lang_matcher(
        self, path: str
    ) -> Tuple[Optional[str], Optional[str]]:
        match = REGEX_STATIC_LANG.match(path)
        if match:
            lang, mname, version, file_name = match.group('l', 'm', 'v', 'f')
            if mname:
                mod = self.app._modules.get(mname)
                spath = mod._static_path if mod else self.app.static_path
            else:
                spath = self.app.static_path
            static_file = os.path.join(spath, file_name)
            if lang:
                lang_file = os.path.join(spath, lang, file_name)
                if os.path.exists(lang_file):
                    static_file = lang_file
            return static_file, version
        return None, None

    def _static_nolang_matcher(
        self, path: str
    ) -> Tuple[Optional[str], Optional[str]]:
        if path.startswith('/static'):
            mname, version, file_name = REGEX_STATIC.match(path).group('m', 'v', 'f')
            if mname:
                mod = self.app._modules.get(mname[2:-3])
                static_file = os.path.join(mod._static_path, file_name) if mod else None
            else:
                static_file = os.path.join(self.app.static_path, file_name)
            return static_file, version
        return None, None

    async def _static_response(self, file_path: str) -> HTTPFile:
        return HTTPFile(file_path)

    def _static_handler(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ) -> Awaitable[HTTPResponse]:
        path = scope['emt.path']
        #: handle internal assets
        if path.startswith('/__emmett__'):
            file_name = path[12:]
            static_file = os.path.join(
                os.path.dirname(__file__), '..', 'assets', file_name)
            if os.path.splitext(static_file)[1] == 'html':
                return self._http_response(404)
            return self._static_response(static_file)
        #: handle app assets
        static_file, _ = self.static_matcher(path)
        if static_file:
            return self._static_response(static_file)
        return self.dynamic_handler(scope, receive, send)

    async def dynamic_handler(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ) -> HTTPResponse:
        ctx = RequestContext(
            self.app,
            scope,
            receive,
            send,
            Request,
            Response
        )
        ctx_token = current._init_(ctx)
        try:
            http = await self.router.dispatch(ctx.request, ctx.response)
        except HTTPResponse as http_exception:
            http = http_exception
            #: render error with handlers if in app
            error_handler = self.app.error_handlers.get(http.status_code)
            if error_handler:
                http = HTTP(
                    http.status_code,
                    await error_handler(),
                    headers=ctx.response.headers,
                    cookies=ctx.response.cookies
                )
        except RequestCancelled:
            raise
        except Exception:
            self.app.log.exception('Application exception:')
            http = HTTP(
                500,
                await self.error_handler(),
                headers=ctx.response.headers
            )
        finally:
            current._close_(ctx_token)
        return http

    async def _debug_handler(self) -> str:
        current.response.headers._data['content-type'] = (
            'text/html; charset=utf-8'
        )
        return debug_handler(smart_traceback(self.app))

    async def _exception_handler(self) -> str:
        current.response.headers._data['content-type'] = 'text/plain'
        return 'Internal error'


class WSHandler(RequestHandler):
    __slots__ = ['pre_handler']

    def _bind_router(self):
        self.router = self.app._router_ws

    def _configure_methods(self):
        self.pre_handler = (
            self._prefix_handler if self.router._prefix_main else
            self.dynamic_handler)

    async def __call__(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ):
        scope['emt.input'] = asyncio.Queue()
        task_events = asyncio.create_task(
            self.handle_events(scope, receive, send))
        task_request = asyncio.create_task(
            self.handle_request(scope, receive, send))
        _, pending = await asyncio.wait(
            [task_request, task_events], return_when=asyncio.FIRST_COMPLETED
        )
        scope['emt._flow_cancel'] = True
        _cancel_tasks(pending)

    @Handler.on_event('websocket.connect')
    async def event_connect(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        event: Event
    ) -> EventLooper:
        return _event_looper

    @Handler.on_event('websocket.disconnect')
    async def event_disconnect(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        event: Event
    ):
        return

    @Handler.on_event('websocket.receive')
    async def event_receive(
        self,
        scope: Scope,
        receive: Receive,
        send: Send,
        event: Event
    ) -> EventLooper:
        await scope['emt.input'].put(event.get('bytes') or event['text'])
        return _event_looper

    async def handle_request(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ):
        scope['emt.path'] = scope['path'] or '/'
        scope['emt._ws_closed'] = False
        try:
            await self.pre_handler(scope, receive, send)
        except HTTPResponse:
            if not scope['emt._ws_closed']:
                await send({'type': 'websocket.close', 'code': 1006})
        except asyncio.CancelledError:
            if not scope.get('emt._flow_cancel', False):
                self.app.log.exception('Application exception:')
        except Exception:
            if not scope['emt._ws_closed']:
                await send({'type': 'websocket.close', 'code': 1006})
            self.app.log.exception('Application exception:')

    def _prefix_handler(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ) -> Awaitable[None]:
        path = scope['emt.path']
        if not path.startswith(self.router._prefix_main):
            raise HTTP(404)
        scope['emt.path'] = path[self.router._prefix_main_len:] or '/'
        return self.dynamic_handler(scope, receive, send)

    async def dynamic_handler(
        self,
        scope: Scope,
        receive: Receive,
        send: Send
    ):
        ctx = WSContext(
            self.app,
            scope,
            scope['emt.input'].get,
            send,
            Websocket
        )
        ctx_token = current._init_(ctx)
        try:
            await self.router.dispatch(ctx.websocket)
        finally:
            if (
                not scope.get('emt._flow_cancel', False) and
                ctx.websocket._accepted
            ):
                await send({'type': 'websocket.close', 'code': 1000})
                scope['emt._ws_closed'] = True
            current._close_(ctx_token)


async def _event_looper(
    handler: Handler,
    scope: Scope,
    receive: Receive,
    send: Send,
    event: Any = None
) -> Tuple[Union[EventHandler, EventHandlerWrapper], Event]:
    event = await receive()
    return handler.get_event_handler(event['type']), event


async def _event_missing(
    handler: Handler,
    scope: Scope,
    receive: Receive,
    send: Send,
    event: Event
):
    raise RuntimeError(f"Event type '{event['type']}' not recognized")


def _cancel_tasks(tasks):
    for task in tasks:
        task.cancel()



# -*- coding: utf-8 -*-
"""
    emmett.routing.urls
    -------------------

    Provides url builder apis.

    :copyright: 2014 Giovanni Barillari
    :license: BSD-3-Clause
"""

from urllib.parse import quote as uquote

from ..ctx import current


class UrlBuilder:
    __slots__ = ('components', '_args')

    def __init__(self, components=[]):
        if not components:
            self.components = ['/{}']
            self._args = ['']
        else:
            self.components = ['{}'] + components[1:]
            self._args = [components[0]]

    @property
    def path(self):
        return self._args[0]

    def arg(self, value):
        if not self.components:
            self.components.append('/{}')
        return self.components.pop(0).format(value)

    def add_prefix(self, args):
        if current.app._router_http._prefix_main:
            self.components.insert(0, '{}')
            args.insert(0, current.app._router_http._prefix_main)

    def add_language(self, args, language):
        if language:
            self.components.insert(0, '/{}')
            args.insert(0, language)

    def path_prefix(self, scheme, host):
        if scheme and host:
            return '{}://{}'.format(scheme, host)
        return ''

    def args(self, args):
        rv = ''
        for arg in args:
            rv += self.arg(arg)
        return rv

    def params(self, params):
        if params:
            return '?' + '&'.join(
                '%s=%s' % (uquote(str(k)), uquote(str(v)))
                for k, v in params.items()
            )
        return ''

    def url(self, scheme, host, language, args, params):
        args = self._args + args
        self.add_language(args, language)
        self.add_prefix(args)
        return (
            f'{self.path_prefix(scheme, host)}{self.args(args)}'
            f'{self.params(params)}'
        )


class HttpUrlBuilder(UrlBuilder):
    __slots__ = ('components', '_args')

    def add_static_versioning(self, args):
        versioning = current.app._router_http.static_versioning
        if (self.path.startswith('/static') and versioning):
            self.components.insert(1, "/_{}")
            args.insert(1, str(versioning))

    def anchor(self, anchor):
        rv = ''
        if anchor:
            if not isinstance(anchor, (list, tuple)):
                anchor = [anchor]
            for element in anchor:
                rv += '#{}'.format(element)
        return rv

    def url(self, scheme, host, language, args, params, anchor):
        args = self._args + args
        self.add_static_versioning(args)
        self.add_language(args, language)
        self.add_prefix(args)
        return (
            f'{self.path_prefix(scheme, host)}{self.args(args)}'
            f'{self.params(params)}{self.anchor(anchor)}'
        )


class Url:
    __slots__ = []
    http_to_ws_schemes = {'http': 'ws', 'https': 'wss'}

    def http(
        self, path, args=[], params={}, anchor=None,
        sign=None, scheme=None, host=None, language=None
    ):
        if not isinstance(args, (list, tuple)):
            args = [args]
        # allow user to use url('static', 'file')
        if path == 'static':
            path = '/static'
        # routes urls with 'dot' notation
        if '/' not in path:
            module = None
            # urls like 'function' refers to same module
            if '.' not in path:
                namespace = current.app.config.url_default_namespace or \
                    current.app.name
                path = namespace + "." + path
            # urls like '.function' refers to main app module
            elif path.startswith('.'):
                if not hasattr(current, 'request'):
                    raise RuntimeError(
                        f'cannot build url("{path}",...) '
                        'without current request'
                    )
                module = current.request.name.rsplit('.', 1)[0]
                path = module + path
            # find correct route
            try:
                url_components = current.app._router_http.routes_out[path]['path']
                url_host = current.app._router_http.routes_out[path]['host']
                builder = HttpUrlBuilder(url_components)
                # try to use the correct hostname
                if url_host is not None:
                    try:
                        if current.request.host != url_host:
                            scheme = current.request.scheme
                            host = url_host
                    except Exception:
                        pass
            except KeyError:
                if path.endswith('.static'):
                    module = module or path.rsplit('.', 1)[0]
                    builder = HttpUrlBuilder([f'/static/__{module}__'])
                else:
                    raise RuntimeError(f'invalid url("{path}",...)')
        # handle classic urls
        else:
            builder = HttpUrlBuilder([path])
        # add language
        lang = None
        if current.app.language_force_on_url:
            if language:
                #: use the given language if is enabled in application
                if language in current.app.languages:
                    lang = language
            else:
                #: try to use the request language if context exists
                if hasattr(current, 'request'):
                    lang = current.request.language
            if lang == current.app.language_default:
                lang = None
        # # add extension (useless??)
        # if extension:
        #     url = url + '.' + extension
        # scheme=True means to use current scheme
        if scheme is True:
            if not hasattr(current, 'request'):
                raise RuntimeError(
                    f'cannot build url("{path}",...) without current request'
                )
            scheme = current.request.scheme
        # add scheme and host
        if scheme:
            if host is None:
                if not hasattr(current, 'request'):
                    raise RuntimeError(
                        f'cannot build url("{path}",...) '
                        'without current request'
                    )
                host = current.request.host
        # add signature
        if sign:
            if '_signature' in params:
                del params['_signature']
            params['_signature'] = sign(
                path, args, params, anchor, scheme, host, language)
        return builder.url(scheme, host, lang, args, params, anchor)

    def ws(
        self, path, args=[], params={}, scheme=None, host=None, language=None
    ):
        if not isinstance(args, (list, tuple)):
            args = [args]
        # routes urls with 'dot' notation
        if '/' not in path:
            # urls like 'function' refers to same module
            if '.' not in path:
                namespace = current.app.config.url_default_namespace or current.app.name
                path = namespace + "." + path
            # urls like '.function' refers to main app module
            elif path.startswith('.'):
                if not hasattr(current, 'request'):
                    raise RuntimeError(
                        f'cannot build url("{path}",...) '
                        'without current request'
                    )
                module = current.request.name.rsplit('.', 1)[0]
                path = module + path
            # find correct route
            try:
                url_components = current.app._router_ws.routes_out[path]['path']
                url_host = current.app._router_ws.routes_out[path]['host']
                builder = UrlBuilder(url_components)
                # try to use the correct hostname
                if url_host is not None:
                    # TODO: remap host
                    try:
                        if current.request.host != url_host:
                            scheme = self.http_to_ws_schemes[current.request.scheme]
                            host = url_host
                    except Exception:
                        pass
            except KeyError:
                raise RuntimeError(f'invalid url("{path}",...)')
        # handle classic urls
        else:
            builder = UrlBuilder([path])
        # add language
        lang = None
        if current.app.language_force_on_url:
            if language:
                #: use the given language if is enabled in application
                if language in current.app.languages:
                    lang = language
            else:
                #: try to use the request language if context exists
                if hasattr(current, 'request'):
                    lang = current.request.language
            if lang == current.app.language_default:
                lang = None
        # scheme=True means to use current scheme
        if scheme is True:
            if not hasattr(current, 'request'):
                raise RuntimeError(
                    f'cannot build url("{path}",...) without current request'
                )
            scheme = self.http_to_ws_schemes[current.request.scheme]
        # add scheme and host
        if scheme:
            if host is None:
                if not hasattr(current, 'request'):
                    raise RuntimeError(
                        f'cannot build url("{path}",...) '
                        'without current request'
                    )
                host = current.request.host
        return builder.url(scheme, host, lang, args, params)

    def __call__(self, *args, **kwargs):
        return self.http(*args, **kwargs)


url = Url()
